# -*- coding: utf-8 -*- #
# Copyright 2018 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for calling the Composer Environments API."""


from apitools.base.py import exceptions as apitools_exceptions
from googlecloudsdk.api_lib.composer import util as api_util
from googlecloudsdk.api_lib.util import exceptions
from googlecloudsdk.calliope import base
from googlecloudsdk.command_lib.composer.flags import CLOUD_RUN_FUNCTIONS_ROUTING_ALPHA
from googlecloudsdk.command_lib.composer.flags import CLOUD_RUN_FUNCTIONS_ROUTING_BETA
from googlecloudsdk.command_lib.composer.flags import CONNECTION_TYPE_FLAG_ALPHA
from googlecloudsdk.command_lib.composer.flags import CONNECTION_TYPE_FLAG_BETA
from googlecloudsdk.command_lib.composer.flags import CONNECTION_TYPE_FLAG_GA
from googlecloudsdk.command_lib.composer.flags import ENVIRONMENT_SIZE_ALPHA
from googlecloudsdk.command_lib.composer.flags import ENVIRONMENT_SIZE_BETA
from googlecloudsdk.command_lib.composer.flags import ENVIRONMENT_SIZE_GA


def GetService(release_track=base.ReleaseTrack.GA):
  return api_util.GetClientInstance(
      release_track).projects_locations_environments


class CreateEnvironmentFlags:
  """Container holding environment creation flag values.

  Attributes:
    node_count: int or None, the number of VMs to create for the environment
    environment_size: str or None, one of small, medium and large.
    labels: dict(str->str), a dict of user-provided resource labels to apply to
      the environment and its downstream resources
    location: str or None, the Compute Engine zone in which to create the
      environment specified as relative resource name.
    machine_type: str or None, the Compute Engine machine type of the VMs to
      create specified as relative resource name.
    network: str or None, the Compute Engine network to which to connect the
      environment specified as relative resource name.
    subnetwork: str or None, the Compute Engine subnetwork to which to connect
      the environment specified as relative resource name.
    network_attachment: str or None, the Compute Engine network attachment that
      is used as PSC Network entry point.
    env_variables: dict(str->str), a dict of user-provided environment variables
      to provide to the Airflow scheduler, worker, and webserver processes.
    airflow_config_overrides: dict(str->str), a dict of user-provided Airflow
      configuration overrides.
    service_account: str or None, the user-provided service account
    oauth_scopes: [str], the user-provided OAuth scopes
    tags: [str], the user-provided networking tags
    disk_size_gb: int, the disk size of node VMs, in GB
    python_version: str or None, major python version to use within created
      environment.
    image_version: str or None, the desired image for created environment in the
      format of 'composer-(version)-airflow-(version)'
    airflow_executor_type: str or None, the airflow executor type to run task
      instances.
    use_ip_aliases: bool or None, create env cluster nodes using alias IPs.
    cluster_secondary_range_name: str or None, the name of secondary range to
      allocate IP addresses to pods in GKE cluster.
    services_secondary_range_name: str or None, the name of the secondary range
      to allocate IP addresses to services in GKE cluster.
    cluster_ipv4_cidr_block: str or None, the IP address range to allocate IP
      adresses to pods in GKE cluster.
    services_ipv4_cidr_block: str or None, the IP address range to allocate IP
      addresses to services in GKE cluster.
    max_pods_per_node: int or None, the maximum number of pods that can be
      assigned to a GKE cluster node.
    enable_ip_masq_agent: bool or None, when enabled, the GKE IP Masq Agent is
      deployed to the cluster.
    enable_private_environment: bool or None, create env cluster nodes with no
      public IP addresses.
    disable_private_environment: bool or None, disable private environment.
    private_endpoint: bool or None, managed env cluster using the private IP
      address of the master API endpoint.
    master_ipv4_cidr: IPv4 CIDR range to use for the cluster master network.
    privately_used_public_ips: bool or None, when enabled, GKE pod and services
      can use IPs from public (non-RFC1918) ranges.
    web_server_ipv4_cidr: IPv4 CIDR range to use for Web Server network.
    cloud_sql_ipv4_cidr: IPv4 CIDR range to use for Cloud SQL network.
    composer_network_ipv4_cidr: IPv4 CIDR range to use for Composer network.
    connection_subnetwork: str or None, the Compute Engine subnetwork from which
      to reserve the IP address for internal connections, specified as relative
      resource name.
    connection_type: str or None, mode of internal connectivity within the Cloud
      Composer environment. Can be VPC_PEERING or PRIVATE_SERVICE_CONNECT.
    web_server_access_control: [{string: string}], List of IP ranges with
      descriptions to allow access to the web server.
    cloud_sql_machine_type: str or None, Cloud SQL machine type used by the
      Airflow database.
    cloud_sql_preferred_zone: str or None, Cloud SQL db preferred zone. Can be
      specified only in Composer 2.0.0.
    web_server_machine_type: str or None, machine type used by the Airflow web
      server
    kms_key: str or None, the user-provided customer-managed encryption key
      resource name
    scheduler_cpu: float or None, CPU allocated to Airflow scheduler. Can be
      specified only in Composer 2.0.0.
    worker_cpu: float or None, CPU allocated to each Airflow worker. Can be
      specified only in Composer 2.0.0.
    web_server_cpu: float or None, CPU allocated to Airflow web server. Can be
      specified only in Composer 2.0.0.
    scheduler_memory_gb: float or None, memory allocated to Airflow scheduler.
      Can be specified only in Composer 2.0.0.
    worker_memory_gb: float or None, memory allocated to each Airflow worker.
      Can be specified only in Composer 2.0.0.
    web_server_memory_gb: float or None, memory allocated to Airflow web server.
      Can be specified only in Composer 2.0.0.
    scheduler_storage_gb: float or None, storage allocated to Airflow scheduler.
      Can be specified only in Composer 2.0.0.
    worker_storage_gb: float or None, storage allocated to each Airflow worker.
      Can be specified only in Composer 2.0.0.
    web_server_storage_gb: float or None, storage allocated to Airflow web
      server. Can be specified only in Composer 2.0.0.
    min_workers: int or None, minimum number of workers in the Environment. Can
      be specified only in Composer 2.0.0.
    max_workers: int or None, maximum number of workers in the Environment. Can
      be specified only in Composer 2.0.0.
    scheduler_count: int or None, number of schedulers in the Environment.
    maintenance_window_start: Datetime or None, the starting time of the
      maintenance window
    maintenance_window_end: Datetime or None, the ending time of the maintenance
      window
    maintenance_window_recurrence: str or None, the recurrence of the
      maintenance window
    enable_master_authorized_networks: bool or None, whether master authorized
      networks should be enabled
    master_authorized_networks: list(str), master authorized networks
    airflow_database_retention_days: Optional[int], the number of retention days
      for airflow database data retention mechanism. Infinite retention will be
      applied in case `0` or no integer is provided.
    release_track: base.ReleaseTrack, the release track of command. Will dictate
      which Composer client library will be used.
    enable_triggerer: bool or None, enable triggerer in the Environment. Can be
      specified only in Airflow 2.2.x and greater
    triggerer_cpu: float or None, CPU allocated to Airflow triggerer. Can be
      specified only in Airflow 2.2.x and greater
    triggerer_count: int or None, number of Airflow triggerers. Can be specified
      only in Airflow 2.2.x and greater
    triggerer_memory_gb: float or None, memory allocated to Airflow triggerer.
      Can be specified only in Airflow 2.2.x and greater
    enable_scheduled_snapshot_creation: bool or None, whether the automatic
      snapshot creation should be enabled
    snapshot_creation_schedule: str or None, cron expression that specifies when
      snapshots will be created
    snapshot_location: str or None, a Cloud Storage location used to store
      automatically created snapshots
    snapshot_schedule_timezone: str or None, time zone that sets the context to
      interpret snapshot_creation_schedule
    enable_cloud_data_lineage_integration: bool or None, whether Cloud Data
      Lineage integration should be enabled
    disable_cloud_data_lineage_integration: bool or None, whether Cloud Data
      Lineage integration should be disabled
    enable_high_resilience: bool or None, whether high resilience should be
      enabled
    enable_logs_in_cloud_logging_only: bool or None, whether logs in cloud
      logging only should be enabled
    disable_logs_in_cloud_logging_only: bool or None, whether logs in cloud
      logging only should be disabled
    support_web_server_plugins: bool or None, whether to enable/disable the
      support for web server plugins
    dag_processor_cpu: float or None, CPU allocated to Airflow dag processor.
      Can be specified only in Composer 3.
    dag_processor_count: int or None, number of Airflow dag processors. Can be
      specified only in Composer 3.
    dag_processor_memory_gb: float or None, memory allocated to Airflow dag
      processor. Can be specified only in Composer 3.
    dag_processor_storage_gb: float or None, storage allocated to Airflow dag
      processor. Can be specified only in Composer 3.
    composer_internal_ipv4_cidr_block: str or None. The IP range in CIDR
      notation to use internally by Cloud Composer. Can be specified only in
      Composer 3.
    enable_private_builds_only: bool or None, whether to enable the support for
      private only builds.
    disable_private_builds_only: bool or None, whether to disable the support
      for private only builds.
    storage_bucket: str or None. An existing Cloud Storage bucket to be used by
      the environment.
    cloud_run_functions_routing: str or None, the routing method for cloud run
      functions. Can be DIRECT or VIA_NETWORK_ATTACHMENT.
    enable_development_mode: bool or None, whether to enable the development
      mode.
  """

  # TODO(b/154131605): This a type that is an immutable data object. Can't use
  # attrs because it's not part of googlecloudsdk and can't use namedtuple
  # because it's not efficient on python 2 (it generates code, which needs
  # to be parsed and interpretted). Remove this code when we get support
  # for attrs or another dumb data object in gcloud.
  def __init__(
      self,
      node_count=None,
      environment_size=None,
      labels=None,
      location=None,
      machine_type=None,
      network=None,
      subnetwork=None,
      network_attachment=None,
      env_variables=None,
      airflow_config_overrides=None,
      service_account=None,
      oauth_scopes=None,
      tags=None,
      disk_size_gb=None,
      python_version=None,
      image_version=None,
      airflow_executor_type=None,
      use_ip_aliases=None,
      cluster_secondary_range_name=None,
      services_secondary_range_name=None,
      cluster_ipv4_cidr_block=None,
      services_ipv4_cidr_block=None,
      max_pods_per_node=None,
      enable_ip_masq_agent=None,
      enable_private_environment=None,
      disable_private_environment=None,
      private_endpoint=None,
      master_ipv4_cidr=None,
      privately_used_public_ips=None,
      web_server_ipv4_cidr=None,
      cloud_sql_ipv4_cidr=None,
      composer_network_ipv4_cidr=None,
      connection_subnetwork=None,
      connection_type=None,
      web_server_access_control=None,
      cloud_sql_machine_type=None,
      web_server_machine_type=None,
      kms_key=None,
      scheduler_cpu=None,
      worker_cpu=None,
      web_server_cpu=None,
      scheduler_memory_gb=None,
      worker_memory_gb=None,
      web_server_memory_gb=None,
      scheduler_storage_gb=None,
      worker_storage_gb=None,
      web_server_storage_gb=None,
      min_workers=None,
      max_workers=None,
      scheduler_count=None,
      maintenance_window_start=None,
      maintenance_window_end=None,
      maintenance_window_recurrence=None,
      enable_master_authorized_networks=None,
      master_authorized_networks=None,
      airflow_database_retention_days=None,
      release_track=base.ReleaseTrack.GA,
      enable_triggerer=None,
      triggerer_cpu=None,
      triggerer_count=None,
      triggerer_memory_gb=None,
      enable_scheduled_snapshot_creation=None,
      snapshot_creation_schedule=None,
      snapshot_location=None,
      snapshot_schedule_timezone=None,
      enable_cloud_data_lineage_integration=None,
      disable_cloud_data_lineage_integration=None,
      enable_high_resilience=None,
      enable_logs_in_cloud_logging_only=None,
      disable_logs_in_cloud_logging_only=None,
      cloud_sql_preferred_zone=None,
      support_web_server_plugins=None,
      dag_processor_cpu=None,
      dag_processor_count=None,
      dag_processor_memory_gb=None,
      dag_processor_storage_gb=None,
      composer_internal_ipv4_cidr_block=None,
      enable_private_builds_only=None,
      disable_private_builds_only=None,
      storage_bucket=None,
      cloud_run_functions_routing=None,
      enable_development_mode=None,
  ):
    self.node_count = node_count
    self.environment_size = environment_size
    self.labels = labels
    self.location = location
    self.machine_type = machine_type
    self.network = network
    self.subnetwork = subnetwork
    self.network_attachment = network_attachment
    self.env_variables = env_variables
    self.airflow_config_overrides = airflow_config_overrides
    self.service_account = service_account
    self.oauth_scopes = oauth_scopes
    self.tags = tags
    self.disk_size_gb = disk_size_gb
    self.python_version = python_version
    self.image_version = image_version
    self.airflow_executor_type = airflow_executor_type
    self.use_ip_aliases = use_ip_aliases
    self.cluster_secondary_range_name = cluster_secondary_range_name
    self.services_secondary_range_name = services_secondary_range_name
    self.cluster_ipv4_cidr_block = cluster_ipv4_cidr_block
    self.services_ipv4_cidr_block = services_ipv4_cidr_block
    self.max_pods_per_node = max_pods_per_node
    self.enable_ip_masq_agent = enable_ip_masq_agent
    self.enable_private_environment = enable_private_environment
    self.disable_private_environment = disable_private_environment
    self.private_endpoint = private_endpoint
    self.master_ipv4_cidr = master_ipv4_cidr
    self.privately_used_public_ips = privately_used_public_ips
    self.web_server_ipv4_cidr = web_server_ipv4_cidr
    self.cloud_sql_ipv4_cidr = cloud_sql_ipv4_cidr
    self.composer_network_ipv4_cidr = composer_network_ipv4_cidr
    self.connection_subnetwork = connection_subnetwork
    self.connection_type = connection_type
    self.web_server_access_control = web_server_access_control
    self.cloud_sql_machine_type = cloud_sql_machine_type
    self.web_server_machine_type = web_server_machine_type
    self.kms_key = kms_key
    self.scheduler_cpu = scheduler_cpu
    self.worker_cpu = worker_cpu
    self.web_server_cpu = web_server_cpu
    self.scheduler_memory_gb = scheduler_memory_gb
    self.worker_memory_gb = worker_memory_gb
    self.web_server_memory_gb = web_server_memory_gb
    self.scheduler_storage_gb = scheduler_storage_gb
    self.worker_storage_gb = worker_storage_gb
    self.web_server_storage_gb = web_server_storage_gb
    self.min_workers = min_workers
    self.max_workers = max_workers
    self.scheduler_count = scheduler_count
    self.enable_triggerer = enable_triggerer
    self.triggerer_cpu = triggerer_cpu
    self.triggerer_count = triggerer_count
    self.triggerer_memory_gb = triggerer_memory_gb
    self.maintenance_window_start = maintenance_window_start
    self.maintenance_window_end = maintenance_window_end
    self.maintenance_window_recurrence = maintenance_window_recurrence
    self.enable_master_authorized_networks = enable_master_authorized_networks
    self.master_authorized_networks = master_authorized_networks
    self.airflow_database_retention_days = airflow_database_retention_days
    self.release_track = release_track
    self.enable_scheduled_snapshot_creation = enable_scheduled_snapshot_creation
    self.snapshot_creation_schedule = snapshot_creation_schedule
    self.snapshot_location = snapshot_location
    self.snapshot_schedule_timezone = snapshot_schedule_timezone
    self.enable_cloud_data_lineage_integration = (
        enable_cloud_data_lineage_integration
    )
    self.disable_cloud_data_lineage_integration = (
        disable_cloud_data_lineage_integration
    )
    self.enable_high_resilience = enable_high_resilience
    self.enable_logs_in_cloud_logging_only = enable_logs_in_cloud_logging_only
    self.disable_logs_in_cloud_logging_only = disable_logs_in_cloud_logging_only
    self.cloud_sql_preferred_zone = cloud_sql_preferred_zone
    self.support_web_server_plugins = support_web_server_plugins
    self.dag_processor_cpu = dag_processor_cpu
    self.dag_processor_storage_gb = dag_processor_storage_gb
    self.dag_processor_memory_gb = dag_processor_memory_gb
    self.dag_processor_count = dag_processor_count
    self.composer_internal_ipv4_cidr_block = composer_internal_ipv4_cidr_block
    self.enable_private_builds_only = enable_private_builds_only
    self.disable_private_builds_only = disable_private_builds_only
    self.storage_bucket = storage_bucket
    self.cloud_run_functions_routing = cloud_run_functions_routing
    self.enable_development_mode = enable_development_mode


def _CreateNodeConfig(messages, flags):
  """Creates node config from parameters, returns None if config is empty."""
  if not (flags.location or flags.machine_type or flags.network or
          flags.subnetwork or flags.service_account or flags.oauth_scopes or
          flags.tags or flags.disk_size_gb or flags.use_ip_aliases or
          flags.cluster_secondary_range_name or flags.network_attachment or
          flags.services_secondary_range_name or flags.cluster_ipv4_cidr_block
          or flags.services_ipv4_cidr_block or flags.enable_ip_masq_agent or
          flags.composer_internal_ipv4_cidr_block or
          flags.cloud_run_functions_routing):
    return None

  config = messages.NodeConfig(
      location=flags.location,
      machineType=flags.machine_type,
      network=flags.network,
      subnetwork=flags.subnetwork,
      serviceAccount=flags.service_account,
      diskSizeGb=flags.disk_size_gb)
  if flags.cloud_run_functions_routing:
    routing_enum = None
    if flags.release_track == base.ReleaseTrack.BETA:
      routing_enum = CLOUD_RUN_FUNCTIONS_ROUTING_BETA.GetEnumForChoice(
          flags.cloud_run_functions_routing
      )
    elif flags.release_track == base.ReleaseTrack.ALPHA:
      routing_enum = CLOUD_RUN_FUNCTIONS_ROUTING_ALPHA.GetEnumForChoice(
          flags.cloud_run_functions_routing
      )
    if routing_enum is not None:
      config.trafficRoutingConfig = messages.TrafficRoutingConfig(
          cloudRunFunctionsRouting=routing_enum
      )
  if flags.network_attachment:
    config.composerNetworkAttachment = flags.network_attachment
  if flags.composer_internal_ipv4_cidr_block:
    config.composerInternalIpv4CidrBlock = (
        flags.composer_internal_ipv4_cidr_block
    )
  if flags.oauth_scopes:
    config.oauthScopes = sorted([s.strip() for s in flags.oauth_scopes])
  if flags.tags:
    config.tags = sorted([t.strip() for t in flags.tags])
  if (flags.use_ip_aliases or flags.cluster_secondary_range_name or
      flags.services_secondary_range_name or flags.cluster_ipv4_cidr_block or
      flags.services_ipv4_cidr_block):
    config.ipAllocationPolicy = messages.IPAllocationPolicy(
        useIpAliases=flags.use_ip_aliases,
        clusterSecondaryRangeName=flags.cluster_secondary_range_name,
        servicesSecondaryRangeName=flags.services_secondary_range_name,
        clusterIpv4CidrBlock=flags.cluster_ipv4_cidr_block,
        servicesIpv4CidrBlock=flags.services_ipv4_cidr_block,
    )

    if flags.max_pods_per_node:
      config.maxPodsPerNode = flags.max_pods_per_node

  if flags.enable_ip_masq_agent:
    config.enableIpMasqAgent = flags.enable_ip_masq_agent
  return config


def _CreateConfig(messages, flags, is_composer_v1):
  """Creates environment config from parameters, returns None if config is empty."""
  node_config = _CreateNodeConfig(messages, flags)
  if not (node_config or flags.node_count or flags.kms_key or
          flags.image_version or flags.env_variables or
          flags.airflow_config_overrides or flags.python_version or
          flags.airflow_executor_type or flags.maintenance_window_start or
          flags.maintenance_window_end or flags.maintenance_window_recurrence or
          flags.enable_private_environment or
          flags.disable_private_environment or
          flags.web_server_access_control or flags.cloud_sql_machine_type or
          flags.web_server_machine_type or flags.scheduler_cpu or
          flags.worker_cpu or flags.web_server_cpu or
          flags.scheduler_memory_gb or flags.worker_memory_gb or
          flags.web_server_memory_gb or flags.scheduler_storage_gb or
          flags.worker_storage_gb or flags.web_server_storage_gb or
          flags.environment_size or flags.min_workers or flags.max_workers or
          flags.scheduler_count or flags.airflow_database_retention_days or
          flags.triggerer_cpu or flags.triggerer_memory or
          flags.enable_triggerer or flags.enable_scheduled_snapshot_creation or
          flags.snapshot_creation_schedule or flags.snapshot_location or
          flags.snapshot_schedule_timezone or
          flags.enable_cloud_data_lineage_integration or
          flags.disable_cloud_data_lineage_integration):
    return None

  config = messages.EnvironmentConfig()
  if flags.node_count:
    config.nodeCount = flags.node_count
  if node_config:
    config.nodeConfig = node_config
  if flags.kms_key:
    config.encryptionConfig = messages.EncryptionConfig(
        kmsKeyName=flags.kms_key)
  if flags.environment_size:
    if flags.release_track == base.ReleaseTrack.GA:
      config.environmentSize = ENVIRONMENT_SIZE_GA.GetEnumForChoice(
          flags.environment_size)
    elif flags.release_track == base.ReleaseTrack.BETA:
      config.environmentSize = ENVIRONMENT_SIZE_BETA.GetEnumForChoice(
          flags.environment_size)
    elif flags.release_track == base.ReleaseTrack.ALPHA:
      config.environmentSize = ENVIRONMENT_SIZE_ALPHA.GetEnumForChoice(
          flags.environment_size)
  if (
      flags.image_version
      or flags.env_variables
      or flags.airflow_config_overrides
      or flags.python_version
      or flags.airflow_executor_type
      or (flags.scheduler_count and is_composer_v1)
      or flags.enable_cloud_data_lineage_integration
      or flags.disable_cloud_data_lineage_integration
  ):
    config.softwareConfig = messages.SoftwareConfig()
    if flags.image_version:
      config.softwareConfig.imageVersion = flags.image_version
    if flags.env_variables:
      config.softwareConfig.envVariables = api_util.DictToMessage(
          flags.env_variables, messages.SoftwareConfig.EnvVariablesValue)
    if flags.airflow_config_overrides:
      config.softwareConfig.airflowConfigOverrides = api_util.DictToMessage(
          flags.airflow_config_overrides,
          messages.SoftwareConfig.AirflowConfigOverridesValue)
    if flags.python_version:
      config.softwareConfig.pythonVersion = flags.python_version
    if flags.airflow_executor_type:
      config.softwareConfig.airflowExecutorType = ConvertToTypeEnum(
          messages.SoftwareConfig.AirflowExecutorTypeValueValuesEnum,
          flags.airflow_executor_type,
      )
    if flags.support_web_server_plugins is not None:
      if flags.support_web_server_plugins:
        config.softwareConfig.webServerPluginsMode = (
            messages.SoftwareConfig.WebServerPluginsModeValueValuesEnum.PLUGINS_ENABLED
        )
      else:
        config.softwareConfig.webServerPluginsMode = (
            messages.SoftwareConfig.WebServerPluginsModeValueValuesEnum.PLUGINS_DISABLED
        )

    if flags.scheduler_count and is_composer_v1:
      config.softwareConfig.schedulerCount = flags.scheduler_count
    if (
        flags.enable_cloud_data_lineage_integration
        or flags.disable_cloud_data_lineage_integration
    ):
      config.softwareConfig.cloudDataLineageIntegration = (
          messages.CloudDataLineageIntegration(
              enabled=(
                  True if flags.enable_cloud_data_lineage_integration else False
              ),
          )
      )

  if flags.maintenance_window_start:
    assert flags.maintenance_window_end, 'maintenance_window_end is missing'
    assert flags.maintenance_window_recurrence, (
        'maintenance_window_recurrence is missing')
    config.maintenanceWindow = messages.MaintenanceWindow(
        startTime=flags.maintenance_window_start.isoformat(),
        endTime=flags.maintenance_window_end.isoformat(),
        recurrence=flags.maintenance_window_recurrence)
  if flags.airflow_database_retention_days is not None:
    if flags.airflow_database_retention_days == 0:
      config.dataRetentionConfig = messages.DataRetentionConfig(
          airflowMetadataRetentionConfig=messages.AirflowMetadataRetentionPolicyConfig(
              retentionMode=messages.AirflowMetadataRetentionPolicyConfig.RetentionModeValueValuesEnum.RETENTION_MODE_DISABLED,
          )
      )
    else:
      config.dataRetentionConfig = messages.DataRetentionConfig(
          airflowMetadataRetentionConfig=messages.AirflowMetadataRetentionPolicyConfig(
              retentionDays=flags.airflow_database_retention_days,
              retentionMode=messages.AirflowMetadataRetentionPolicyConfig.RetentionModeValueValuesEnum.RETENTION_MODE_ENABLED,
          )
      )

  if flags.enable_scheduled_snapshot_creation:
    config.recoveryConfig = messages.RecoveryConfig(
        scheduledSnapshotsConfig=messages.ScheduledSnapshotsConfig(
            enabled=flags.enable_scheduled_snapshot_creation,
            snapshotCreationSchedule=flags.snapshot_creation_schedule,
            snapshotLocation=flags.snapshot_location,
            timeZone=flags.snapshot_schedule_timezone))

  if (
      flags.enable_private_environment or flags.disable_private_environment
      or flags.enable_private_builds_only
      or flags.disable_private_builds_only
  ):
    # Adds a PrivateClusterConfig, if necessary.
    private_cluster_config = None
    networking_config = None
    if flags.private_endpoint or flags.master_ipv4_cidr:
      private_cluster_config = messages.PrivateClusterConfig(
          enablePrivateEndpoint=flags.private_endpoint,
          masterIpv4CidrBlock=flags.master_ipv4_cidr)
    if flags.connection_type:
      if flags.release_track == base.ReleaseTrack.GA:
        connection_type = CONNECTION_TYPE_FLAG_GA.GetEnumForChoice(
            flags.connection_type)
      elif flags.release_track == base.ReleaseTrack.BETA:
        connection_type = CONNECTION_TYPE_FLAG_BETA.GetEnumForChoice(
            flags.connection_type)
      elif flags.release_track == base.ReleaseTrack.ALPHA:
        connection_type = CONNECTION_TYPE_FLAG_ALPHA.GetEnumForChoice(
            flags.connection_type)
      networking_config = messages.NetworkingConfig(
          connectionType=connection_type)

    if flags.enable_private_environment:
      networking_type = (
          messages.PrivateEnvironmentConfig.NetworkingTypeValueValuesEnum.PRIVATE
      )
    elif flags.disable_private_environment:
      networking_type = (
          messages.PrivateEnvironmentConfig.NetworkingTypeValueValuesEnum.PUBLIC
      )
    else:
      networking_type = None

    private_env_config_args = {
        'networkingType': networking_type,
        'privateClusterConfig': private_cluster_config,
        'networkingConfig': networking_config,
    }

    if flags.web_server_ipv4_cidr is not None:
      private_env_config_args[
          'webServerIpv4CidrBlock'] = flags.web_server_ipv4_cidr
    if flags.cloud_sql_ipv4_cidr is not None:
      private_env_config_args[
          'cloudSqlIpv4CidrBlock'] = flags.cloud_sql_ipv4_cidr
    if flags.composer_network_ipv4_cidr is not None:
      private_env_config_args[
          'cloudComposerNetworkIpv4CidrBlock'] = flags.composer_network_ipv4_cidr
    if flags.privately_used_public_ips is not None:
      private_env_config_args[
          'enablePrivatelyUsedPublicIps'] = flags.privately_used_public_ips
    if flags.connection_subnetwork is not None:
      private_env_config_args[
          'cloudComposerConnectionSubnetwork'] = flags.connection_subnetwork
    if flags.enable_private_builds_only or flags.disable_private_builds_only:
      private_env_config_args['enablePrivateBuildsOnly'] = (
          True if flags.enable_private_builds_only else False
      )
    config.privateEnvironmentConfig = messages.PrivateEnvironmentConfig(
        **private_env_config_args)

  # Builds webServerNetworkAccessControl, if necessary.
  if flags.web_server_access_control is not None:
    config.webServerNetworkAccessControl = BuildWebServerNetworkAccessControl(
        flags.web_server_access_control, flags.release_track)

  if flags.enable_high_resilience:
    config.resilienceMode = (
        messages.EnvironmentConfig.ResilienceModeValueValuesEnum.HIGH_RESILIENCE
    )
  if flags.enable_logs_in_cloud_logging_only:
    task_logs_retention_config = messages.TaskLogsRetentionConfig(
        storageMode=messages.TaskLogsRetentionConfig.StorageModeValueValuesEnum.CLOUD_LOGGING_ONLY
    )
    config.dataRetentionConfig = messages.DataRetentionConfig(
        taskLogsRetentionConfig=task_logs_retention_config
    )
  if flags.disable_logs_in_cloud_logging_only:
    task_logs_retention_config = messages.TaskLogsRetentionConfig(
        storageMode=messages.TaskLogsRetentionConfig.StorageModeValueValuesEnum.CLOUD_LOGGING_AND_CLOUD_STORAGE
    )
    config.dataRetentionConfig = messages.DataRetentionConfig(
        taskLogsRetentionConfig=task_logs_retention_config
    )
  if flags.cloud_sql_machine_type:
    config.databaseConfig = messages.DatabaseConfig(
        machineType=flags.cloud_sql_machine_type)
  if flags.cloud_sql_preferred_zone:
    config.databaseConfig = messages.DatabaseConfig(
        zone=flags.cloud_sql_preferred_zone
    )
  if flags.web_server_machine_type:
    config.webServerConfig = messages.WebServerConfig(
        machineType=flags.web_server_machine_type)

  if flags.enable_master_authorized_networks:
    networks = flags.master_authorized_networks if flags.master_authorized_networks else []
    config.masterAuthorizedNetworksConfig = messages.MasterAuthorizedNetworksConfig(
        enabled=True,
        cidrBlocks=[
            messages.CidrBlock(cidrBlock=network) for network in networks
        ])

  composer_v2_flags = [
      flags.scheduler_cpu,
      flags.worker_cpu,
      flags.web_server_cpu,
      flags.scheduler_memory_gb,
      flags.worker_memory_gb,
      flags.web_server_memory_gb,
      flags.scheduler_storage_gb,
      flags.worker_storage_gb,
      flags.web_server_storage_gb,
      flags.min_workers,
      flags.max_workers,
      flags.triggerer_memory_gb,
      flags.triggerer_cpu,
      flags.enable_triggerer,
      flags.triggerer_count,
      flags.dag_processor_cpu,
      flags.dag_processor_count,
      flags.dag_processor_memory_gb,
      flags.dag_processor_storage_gb,
  ]
  composer_v2_flag_used = any(flag is not None for flag in composer_v2_flags)
  if composer_v2_flag_used or (flags.scheduler_count and not is_composer_v1):
    config.workloadsConfig = _CreateWorkloadConfig(messages, flags)
  return config


def _CreateWorkloadConfig(messages, flags):
  """Creates workload config from parameters."""
  workload_resources = dict(
      scheduler=messages.SchedulerResource(
          cpu=flags.scheduler_cpu,
          memoryGb=flags.scheduler_memory_gb,
          storageGb=flags.scheduler_storage_gb,
          count=flags.scheduler_count),
      webServer=messages.WebServerResource(
          cpu=flags.web_server_cpu,
          memoryGb=flags.web_server_memory_gb,
          storageGb=flags.web_server_storage_gb),
      worker=messages.WorkerResource(
          cpu=flags.worker_cpu,
          memoryGb=flags.worker_memory_gb,
          storageGb=flags.worker_storage_gb,
          minCount=flags.min_workers,
          maxCount=flags.max_workers))
  if (
      flags.enable_triggerer
      or flags.triggerer_cpu
      or flags.triggerer_memory_gb
      or flags.triggerer_count is not None
  ):
    triggerer_count = 1 if flags.enable_triggerer else 0
    if flags.triggerer_count is not None:
      triggerer_count = flags.triggerer_count
    workload_resources['triggerer'] = messages.TriggererResource(
        cpu=flags.triggerer_cpu,
        memoryGb=flags.triggerer_memory_gb,
        count=triggerer_count
    )
  if (
      flags.dag_processor_cpu
      or flags.dag_processor_count is not None
      or flags.dag_processor_memory_gb
      or flags.dag_processor_storage_gb
  ):
    workload_resources['dagProcessor'] = messages.DagProcessorResource(
        cpu=flags.dag_processor_cpu,
        memoryGb=flags.dag_processor_memory_gb,
        storageGb=flags.dag_processor_storage_gb,
        count=flags.dag_processor_count,
    )

  return messages.WorkloadsConfig(**workload_resources)


def Create(environment_ref, flags, is_composer_v1):
  """Calls the Composer Environments.Create method.

  Args:
    environment_ref: Resource, the Composer environment resource to create.
    flags: CreateEnvironmentFlags, the flags provided for environment creation.
    is_composer_v1: boolean representing if creation request is for Composer
      1.*.* image versions.

  Returns:
    Operation: the operation corresponding to the creation of the environment
  """
  messages = api_util.GetMessagesModule(release_track=flags.release_track)
  # Builds environment message and attaches the configuration
  environment = messages.Environment(name=environment_ref.RelativeName())
  environment.config = _CreateConfig(messages, flags, is_composer_v1)

  if flags.labels:
    environment.labels = api_util.DictToMessage(
        flags.labels, messages.Environment.LabelsValue)

  if flags.storage_bucket:
    environment.storageConfig = messages.StorageConfig(
        bucket=flags.storage_bucket
    )

  if flags.enable_development_mode:
    mode_enum = getattr(messages.Environment, 'ModeValueValuesEnum', None)
    if mode_enum:
      environment.mode = mode_enum.DEVELOPMENT

  try:
    return GetService(release_track=flags.release_track).Create(
        api_util.GetMessagesModule(release_track=flags.release_track)
        .ComposerProjectsLocationsEnvironmentsCreateRequest(
            environment=environment,
            parent=environment_ref.Parent().RelativeName()))
  except apitools_exceptions.HttpForbiddenError as e:
    raise exceptions.HttpException(
        e,
        error_format=(
            'Creation operation failed because of lack of proper '
            'permissions. Please, refer to '
            'https://cloud.google.com/composer/docs/how-to/managing/creating '
            'and Composer Creation Troubleshooting pages to resolve this issue.'
        ))


def ConvertToTypeEnum(type_enum, airflow_executor_type):
  """Converts airflow executor type string to enum.

  Args:
    type_enum: AirflowExecutorTypeValueValuesEnum, executor type enum value.
    airflow_executor_type: string, executor type string value.

  Returns:
    AirflowExecutorTypeValueValuesEnum: the executor type enum value.
  """
  return type_enum(airflow_executor_type)


def Delete(environment_ref, release_track=base.ReleaseTrack.GA):
  """Calls the Composer Environments.Delete method.

  Args:
    environment_ref: Resource, the Composer environment resource to delete.
    release_track: base.ReleaseTrack, the release track of command. Will dictate
      which Composer client library will be used.

  Returns:
    Operation: the operation corresponding to the deletion of the environment
  """
  return GetService(release_track=release_track).Delete(
      api_util.GetMessagesModule(release_track=release_track)
      .ComposerProjectsLocationsEnvironmentsDeleteRequest(
          name=environment_ref.RelativeName()))


def RestartWebServer(environment_ref, release_track=base.ReleaseTrack.GA):
  """Calls the Composer Environments.RestartWebServer method.

  Args:
    environment_ref: Resource, the Composer environment resource to restart the
      web server for.
    release_track: base.ReleaseTrack, the release track of command. Will dictate
      which Composer client library will be used.

  Returns:
    Operation: the operation corresponding to the restart of the web server
  """
  message_module = api_util.GetMessagesModule(release_track=release_track)
  request_message = message_module.ComposerProjectsLocationsEnvironmentsRestartWebServerRequest(
      name=environment_ref.RelativeName())
  return GetService(
      release_track=release_track).RestartWebServer(request_message)


def Hibernate(environment_ref, release_track=base.ReleaseTrack.ALPHA):
  """Calls the Composer Environments.Hibernate method.

  Args:
    environment_ref: Resource, the Composer environment resource to hibernate.
    release_track: base.ReleaseTrack, the release track of command. Will dictate
      which Composer client library will be used.

  Returns:
    Operation: the operation corresponding to hibernating the environment.
  """
  message_module = api_util.GetMessagesModule(release_track=release_track)
  request_message = (
      message_module.ComposerProjectsLocationsEnvironmentsHibernateRequest(
          name=environment_ref.RelativeName()
      )
  )
  return GetService(release_track=release_track).Hibernate(request_message)


def Resume(environment_ref, release_track=base.ReleaseTrack.ALPHA):
  """Calls the Composer Environments.Resume method.

  Args:
    environment_ref: Resource, the Composer environment resource to resume.
    release_track: base.ReleaseTrack, the release track of command. Will dictate
      which Composer client library will be used.

  Returns:
    Operation: the operation corresponding to resuming the environment.
  """
  message_module = api_util.GetMessagesModule(release_track=release_track)
  request_message = (
      message_module.ComposerProjectsLocationsEnvironmentsResumeRequest(
          name=environment_ref.RelativeName()
      )
  )
  return GetService(release_track=release_track).Resume(request_message)


def SaveSnapshot(environment_ref,
                 snapshot_location,
                 release_track=base.ReleaseTrack.ALPHA):
  """Calls the Composer Environments.SaveSnapshot method.

  Args:
    environment_ref: Resource, the Composer environment resource to save the
      snapshot for.
    snapshot_location: location to save the snapshot of the environment.
    release_track: base.ReleaseTrack, the release track of command. Will dictate
      which Composer client library will be used.

  Returns:
    Operation: the operation corresponding to saving the snapshot.
  """
  message_module = api_util.GetMessagesModule(release_track=release_track)
  request_message = message_module.ComposerProjectsLocationsEnvironmentsSaveSnapshotRequest(
      environment=environment_ref.RelativeName(),
      saveSnapshotRequest=message_module.SaveSnapshotRequest(
          snapshotLocation=snapshot_location))
  return GetService(release_track=release_track).SaveSnapshot(request_message)


def LoadSnapshot(environment_ref,
                 skip_pypi_packages_installation,
                 skip_environment_variables_setting,
                 skip_airflow_overrides_setting,
                 skip_gcs_data_copying,
                 snapshot_path,
                 release_track=base.ReleaseTrack.ALPHA):
  """Calls the Composer Environments.LoadSnapshot method.

  Args:
    environment_ref: Resource, the Composer environment resource to Load the
      snapshot for.
    skip_pypi_packages_installation: skip installing the pypi packages during
      the operation.
    skip_environment_variables_setting: skip setting environment variables
      during the operation.
    skip_airflow_overrides_setting: skip setting Airflow overrides during the
      operation.
    skip_gcs_data_copying: skip copying GCS data during the operation.
    snapshot_path: path of the specific snapshot to load the snapshot.
    release_track: base.ReleaseTrack, the release track of command. Will dictate
      which Composer client library will be used.

  Returns:
    Operation: the operation corresponding to loading the snapshot.
  """
  message_module = api_util.GetMessagesModule(release_track=release_track)
  request_message = message_module.ComposerProjectsLocationsEnvironmentsLoadSnapshotRequest(
      environment=environment_ref.RelativeName(),
      loadSnapshotRequest=message_module.LoadSnapshotRequest(
          skipPypiPackagesInstallation=skip_pypi_packages_installation,
          skipEnvironmentVariablesSetting=skip_environment_variables_setting,
          skipAirflowOverridesSetting=skip_airflow_overrides_setting,
          skipGcsDataCopying=skip_gcs_data_copying,
          snapshotPath=snapshot_path))
  return GetService(release_track=release_track).LoadSnapshot(request_message)


def ExecuteAirflowCommand(
    command,
    subcommand,
    parameters,
    environment_ref,
    release_track=base.ReleaseTrack.ALPHA,
):
  """Starts execution of an Airflow CLI command through Composer API.

  Args:
    command: string, the command to execute.
    subcommand: string, the subcommand to execute.
    parameters: string[], optional, additinal parameters for the command.
    environment_ref: Resource, the Composer environment to execute the command.
    release_track: base.ReleaseTrack, the release track of command. Determines
      which Composer client library is used.

  Returns:
    ExecuteAirflowCommandResponse: information about the execution.
  """
  message_module = api_util.GetMessagesModule(release_track=release_track)
  request_message = message_module.ComposerProjectsLocationsEnvironmentsExecuteAirflowCommandRequest(
      environment=environment_ref.RelativeName(),
      executeAirflowCommandRequest=message_module.ExecuteAirflowCommandRequest(
          command=command,
          subcommand=subcommand,
          parameters=parameters,
      ),
  )
  return GetService(release_track=release_track).ExecuteAirflowCommand(
      request_message
  )


def StopAirflowCommand(
    execution_id,
    pod_name,
    pod_namespace,
    force,
    environment_ref,
    release_track=base.ReleaseTrack.ALPHA,
):
  """Stops the execution of an Airflow CLI command.

  Args:
    execution_id: string, the unique ID of execution.
    pod_name: string, the pod the execution is running on.
    pod_namespace: string, the pod's namespace.
    force: boolean, If true, the execution is terminated forcefully (SIGKILL).
      If false, the  execution is stopped gracefully, giving it time for
      cleanup.
    environment_ref: Resource, the Composer environment to stop the command.
    release_track: base.ReleaseTrack, the release track of command. Determines
      which Composer client library is used.

  Returns:
    StopAirflowCommandResponse: information whether stopping the execution was
    successful.
  """
  message_module = api_util.GetMessagesModule(release_track=release_track)
  request_message = message_module.ComposerProjectsLocationsEnvironmentsStopAirflowCommandRequest(
      environment=environment_ref.RelativeName(),
      stopAirflowCommandRequest=message_module.StopAirflowCommandRequest(
          executionId=execution_id,
          pod=pod_name,
          podNamespace=pod_namespace,
          force=force,
      ),
  )
  return GetService(release_track=release_track).StopAirflowCommand(
      request_message
  )


def PollAirflowCommand(
    execution_id,
    pod_name,
    pod_namespace,
    next_line_number,
    environment_ref,
    release_track=base.ReleaseTrack.ALPHA,
):
  """Polls the execution of an Airflow CLI command through Composer API.

  Args:
    execution_id: string, the unique ID of execution.
    pod_name: string, the pod the execution is running on.
    pod_namespace: string, the pod's namespace.
    next_line_number: int, line of the output which should be fetched.
    environment_ref: Resource, the Composer environment to poll the command.
    release_track: base.ReleaseTrack, the release track of command. Determines
      which Composer client library is used.

  Returns:
    PollAirflowCommandResponse: the next output lines from the execution and
    information whether the execution is still running.
  """
  message_module = api_util.GetMessagesModule(release_track=release_track)
  request_message = message_module.ComposerProjectsLocationsEnvironmentsPollAirflowCommandRequest(
      environment=environment_ref.RelativeName(),
      pollAirflowCommandRequest=message_module.PollAirflowCommandRequest(
          executionId=execution_id,
          pod=pod_name,
          podNamespace=pod_namespace,
          nextLineNumber=next_line_number,
      ),
  )
  return GetService(release_track=release_track).PollAirflowCommand(
      request_message
  )


def DatabaseFailover(environment_ref, release_track=base.ReleaseTrack.ALPHA):
  """Triggers the database failover (only for highly resilient environments).

  Args:
    environment_ref: Resource, the Composer environment resource to trigger the
      database failover for.
    release_track: base.ReleaseTrack, the release track of command. Determines
      which Composer client library is used.

  Returns:
    Operation: the operation corresponding to triggering a database failover.
  """
  message_module = api_util.GetMessagesModule(release_track=release_track)
  request_message = message_module.ComposerProjectsLocationsEnvironmentsDatabaseFailoverRequest(
      environment=environment_ref.RelativeName(),
      databaseFailoverRequest=message_module.DatabaseFailoverRequest())
  return GetService(release_track=release_track).DatabaseFailover(
      request_message
  )


def FetchDatabaseProperties(environment_ref,
                            release_track=base.ReleaseTrack.ALPHA):
  """Fetch database properties.

  Args:
    environment_ref: Resource, the Composer environment resource to fetch the
      database properties for.
    release_track: base.ReleaseTrack, the release track of command. Determines
      which Composer client library is used.

  Returns:
    DatabaseProperties: database properties for a given environment.
  """
  message_module = api_util.GetMessagesModule(release_track=release_track)
  request_message = message_module.ComposerProjectsLocationsEnvironmentsFetchDatabasePropertiesRequest(
      environment=environment_ref.RelativeName())
  return GetService(release_track=release_track).FetchDatabaseProperties(
      request_message
  )


def CheckUpgrade(environment_ref,
                 image_version,
                 release_track=base.ReleaseTrack.GA):
  """Calls the Composer Environments.CheckUpgrade method.

  Args:
    environment_ref: Resource, the Composer environment resource to check
      upgrade for.
    image_version: Image version to upgrade to.
    release_track: base.ReleaseTrack, the release track of command. Determines
      which Composer client library is used.

  Returns:
    Operation: the operation corresponding to the upgrade check
  """
  message_module = api_util.GetMessagesModule(release_track=release_track)
  request_message = message_module.ComposerProjectsLocationsEnvironmentsCheckUpgradeRequest(
      environment=environment_ref.RelativeName(),
      checkUpgradeRequest=message_module.CheckUpgradeRequest(
          imageVersion=image_version))
  return GetService(release_track=release_track).CheckUpgrade(request_message)


def Migrate(
    environment_ref,
    request,
    release_track=base.ReleaseTrack.BETA,
):
  """Calls the Composer Environments.Migrate method.

  Args:
    environment_ref: Resource, the Composer environment resource to migrate.
    request: MigrateEnvironmentRequest.
    release_track: base.ReleaseTrack, the release track of command. Determines
      which Composer client library is used.

  Returns:
    Operation: the operation corresponding to the migration.
  """
  message_module = api_util.GetMessagesModule(release_track=release_track)
  request_message = (
      message_module.ComposerProjectsLocationsEnvironmentsMigrateRequest(
          name=environment_ref.RelativeName(),
          migrateEnvironmentRequest=request))
  return GetService(release_track=release_track).Migrate(request_message)


def Get(environment_ref, release_track=base.ReleaseTrack.GA):
  """Calls the Composer Environments.Get method.

  Args:
    environment_ref: Resource, the Composer environment resource to retrieve.
    release_track: base.ReleaseTrack, the release track of command. Will dictate
      which Composer client library will be used.

  Returns:
    Environment: the requested environment
  """
  return GetService(release_track=release_track).Get(
      api_util.GetMessagesModule(release_track=release_track)
      .ComposerProjectsLocationsEnvironmentsGetRequest(
          name=environment_ref.RelativeName()))


def List(location_refs,
         page_size,
         limit=None,
         release_track=base.ReleaseTrack.GA):
  """Lists Composer Environments across all locations.

  Uses a hardcoded list of locations, as there is no way to dynamically
  discover the list of supported locations. Support for new locations
  will be aligned with Cloud SDK releases.

  Args:
    location_refs: [core.resources.Resource], a list of resource reference to
      locations in which to list environments.
    page_size: An integer specifying the maximum number of resources to be
      returned in a single list call.
    limit: An integer specifying the maximum number of environments to list.
      None if all available environments should be returned.
    release_track: base.ReleaseTrack, the release track of command. Will dictate
      which Composer client library will be used.

  Returns:
    list: a generator over Environments in the locations in `location_refs`
  """
  return api_util.AggregateListResults(
      api_util.GetMessagesModule(release_track=release_track)
      .ComposerProjectsLocationsEnvironmentsListRequest,
      GetService(release_track=release_track),
      location_refs,
      'environments',
      page_size,
      limit=limit)


def Patch(environment_ref,
          environment_patch,
          update_mask,
          release_track=base.ReleaseTrack.GA):
  """Calls the Composer Environments.Update method.

  Args:
    environment_ref: Resource, the Composer environment resource to update.
    environment_patch: The Environment message specifying the patch associated
      with the update_mask.
    update_mask: A field mask defining the patch.
    release_track: base.ReleaseTrack, the release track of command. Will dictate
      which Composer client library will be used.

  Returns:
    Operation: the operation corresponding to the environment update
  """
  try:
    return GetService(release_track=release_track).Patch(
        api_util.GetMessagesModule(release_track=release_track)
        .ComposerProjectsLocationsEnvironmentsPatchRequest(
            name=environment_ref.RelativeName(),
            environment=environment_patch,
            updateMask=update_mask))
  except apitools_exceptions.HttpForbiddenError as e:
    raise exceptions.HttpException(
        e,
        error_format=(
            'Update operation failed because of lack of proper '
            'permissions. Please, refer to '
            'https://cloud.google.com/composer/docs/how-to/managing/updating '
            'and Composer Update Troubleshooting pages to resolve this issue.'))


def BuildWebServerNetworkAccessControl(web_server_access_control,
                                       release_track):
  """Builds a WebServerNetworkAccessControl proto given an IP range list.

  If the list is empty, the returned policy is set to ALLOW by default.
  Otherwise, the default policy is DENY with a list of ALLOW rules for each
  of the IP ranges.

  Args:
    web_server_access_control: [{string: string}], list of IP ranges with
      descriptions.
    release_track: base.ReleaseTrack, the release track of command. Will dictate
      which Composer client library will be used.

  Returns:
    WebServerNetworkAccessControl: proto to be sent to the API.
  """
  messages = api_util.GetMessagesModule(release_track=release_track)
  return messages.WebServerNetworkAccessControl(allowedIpRanges=[
      messages.AllowedIpRange(
          value=ip_range['ip_range'], description=ip_range.get('description'))
      for ip_range in web_server_access_control
  ])


def BuildWebServerAllowedIps(allowed_ip_list, allow_all, deny_all):
  """Returns the list of IP ranges that will be sent to the API.

  The resulting IP range list is determined by the options specified in
  environment create or update flags.

  Args:
    allowed_ip_list: [{string: string}], list of IP ranges with descriptions.
    allow_all: bool, True if allow all flag was set.
    deny_all: bool, True if deny all flag was set.

  Returns:
    [{string: string}]: list of IP ranges that will be sent to the API, taking
        into account the values of allow all and deny all flags.
  """
  if deny_all:
    return []
  if allow_all:
    return [{
        'ip_range': '0.0.0.0/0',
        'description': 'Allows access from all IPv4 addresses (default value)',
    }, {
        'ip_range': '::0/0',
        'description': 'Allows access from all IPv6 addresses (default value)',
    }]
  return allowed_ip_list


def DiskSizeBytesToGB(disk_size):
  """Returns a disk size value in GB.

  Args:
    disk_size: int, size in bytes, or None for default value

  Returns:
    int, size in GB
  """
  return disk_size >> 30 if disk_size else disk_size


def MemorySizeBytesToGB(memory_size):
  """Returns a memory size value in GB.

  Args:
    memory_size: int, size in bytes, or None for default value

  Returns:
    float, size in GB rounded to 3 decimal places
  """
  if not memory_size:
    return memory_size
  return round(memory_size / float(1 << 30), 3)



# -*- coding: utf-8 -*- #
# Copyright 2016 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""bigtable instances create command."""


import textwrap

from googlecloudsdk.api_lib.bigtable import clusters
from googlecloudsdk.api_lib.bigtable import util
from googlecloudsdk.calliope import base
from googlecloudsdk.calliope import exceptions
from googlecloudsdk.command_lib.bigtable import arguments
from googlecloudsdk.core import log
from googlecloudsdk.core import resources


@base.UniverseCompatible
@base.ReleaseTracks(base.ReleaseTrack.GA, base.ReleaseTrack.BETA)
class CreateInstance(base.CreateCommand):
  """Create a new Bigtable instance."""

  detailed_help = {
      'EXAMPLES': textwrap.dedent("""\
          To create an instance with id `my-instance-id` with a cluster located
          in `us-east1-c`, run:

            $ {command} my-instance-id --display-name="My Instance" --cluster-config=id=my-cluster-id,zone=us-east1-c

          To create an instance with multiple clusters, run:

            $ {command} my-instance-id --display-name="My Instance" --cluster-config=id=my-cluster-id-1,zone=us-east1-c --cluster-config=id=my-cluster-id-2,zone=us-west1-c,nodes=3

          To create an instance with `HDD` storage and `10` nodes, run:

            $ {command} my-hdd-instance --display-name="HDD Instance" --cluster-storage-type=HDD --cluster-config=id=my-cluster-id,zone=us-east1-c,nodes=10

          To create an instance with edition `ENTERPRISE_PLUS`, run:

            $ {command} my-enterprise-plus-instance --display-name="Enterprise Plus Instance" --edition=ENTERPRISE_PLUS --cluster-config=id=my-cluster-id,zone=us-east1-c,nodes=1
          """),
  }

  @classmethod
  def Args(cls, parser):
    """Register flags for this command."""
    (
        arguments.ArgAdder(parser)
        .AddInstanceDisplayName(required=True)
        .AddClusterConfig()
        .AddDeprecatedCluster()
        .AddDeprecatedClusterZone()
        .AddDeprecatedClusterNodes()
        .AddClusterStorage()
        .AddAsync()
        .AddDeprecatedInstanceType()
        .AddInstanceEdition()
        .AddTags()
    )

    arguments.AddInstanceResourceArg(parser, 'to create', positional=True)
    parser.display_info.AddCacheUpdater(arguments.InstanceCompleter)

  def Run(self, args):
    """Executes the instances create command.

    Args:
      args: an argparse namespace. All the arguments that were provided to this
        command invocation.

    Returns:
      Some value that we want to have printed later.
    """

    return self._Run(args)

  def _Run(self, args):
    """Implements Run() with different possible features flags."""

    cli = util.GetAdminClient()
    ref = args.CONCEPTS.instance.Parse()
    # TODO(b/153576330): This is a workaround for inconsistent collection names.
    parent_ref = resources.REGISTRY.Create(
        'bigtableadmin.projects', projectId=ref.projectsId)
    msgs = util.GetAdminMessages()
    instance_type = msgs.Instance.TypeValueValuesEnum(args.instance_type)

    new_clusters = self._Clusters(args)
    clusters_properties = []
    for cluster_id, cluster in sorted(new_clusters.items()):
      clusters_properties.append(
          msgs.CreateInstanceRequest.ClustersValue.AdditionalProperty(
              key=cluster_id, value=cluster))

    edition = (
        msgs.Instance.EditionValueValuesEnum(args.edition)
        if args.edition is not None
        else None
    )

    msg = msgs.CreateInstanceRequest(
        instanceId=ref.Name(),
        parent=parent_ref.RelativeName(),
        instance=msgs.Instance(
            displayName=args.display_name,
            type=instance_type,
            tags=self._Tags(args),
            edition=edition,
        ),
        clusters=msgs.CreateInstanceRequest.ClustersValue(
            additionalProperties=clusters_properties
        ),
    )
    result = cli.projects_instances.Create(msg)
    operation_ref = util.GetOperationRef(result)

    if args.async_:
      log.CreatedResource(
          operation_ref.RelativeName(),
          kind='bigtable instance {0}'.format(ref.Name()),
          is_async=True)
      return result

    return util.AwaitInstance(
        operation_ref, 'Creating bigtable instance {0}'.format(ref.Name()))

  def _Clusters(self, args):
    """Get the clusters configs from command arguments.

    Args:
      args: the argparse namespace from Run().

    Returns:
      A dict mapping from cluster id to msg.Cluster.
    """

    msgs = util.GetAdminMessages()
    storage_type = msgs.Cluster.DefaultStorageTypeValueValuesEnum(
        args.cluster_storage_type.upper())

    if args.cluster_config is not None:
      if (args.cluster is not None
          or args.cluster_zone is not None
          or args.cluster_num_nodes is not None):
        raise exceptions.InvalidArgumentException(
            '--cluster-config --cluster --cluster-zone --cluster-num-nodes',
            'Use --cluster-config or the combination of --cluster, '
            '--cluster-zone and --cluster-num-nodes to specify cluster(s), not '
            'both.')

      self._ValidateClusterConfigArgs(args.cluster_config)
      new_clusters = {}
      for cluster_dict in args.cluster_config:
        nodes = cluster_dict.get('nodes', 1)
        node_scaling_factor = cluster_dict.get(
            'node-scaling-factor',
            msgs.Cluster.NodeScalingFactorValueValuesEnum(
                msgs.Cluster.NodeScalingFactorValueValuesEnum.NODE_SCALING_FACTOR_1X
            ),
        )
        cluster = msgs.Cluster(
            serveNodes=nodes,
            nodeScalingFactor=node_scaling_factor,
            defaultStorageType=storage_type,
            # TODO(b/36049938): switch location to resource
            # when b/29566669 is fixed on API
            location=util.LocationUrl(cluster_dict['zone']))
        if 'kms-key' in cluster_dict:
          cluster.encryptionConfig = msgs.EncryptionConfig(
              kmsKeyName=cluster_dict['kms-key'])

        if ('autoscaling-min-nodes' in cluster_dict or
            'autoscaling-max-nodes' in cluster_dict or
            'autoscaling-cpu-target' in cluster_dict):
          # autoscaling-storage-target is optional.
          if 'autoscaling-storage-target' in cluster_dict:
            storage_target = cluster_dict['autoscaling-storage-target']
          else:
            storage_target = None

          cluster.clusterConfig = clusters.BuildClusterConfig(
              autoscaling_min=cluster_dict['autoscaling-min-nodes'],
              autoscaling_max=cluster_dict['autoscaling-max-nodes'],
              autoscaling_cpu_target=cluster_dict['autoscaling-cpu-target'],
              autoscaling_storage_target=storage_target)
          # serveNodes must be set to None or 0 to enable Autoscaling.
          # go/cbt-autoscaler-api
          cluster.serveNodes = None

        new_clusters[cluster_dict['id']] = cluster
      return new_clusters
    elif args.cluster is not None:
      if args.cluster_zone is None:
        raise exceptions.InvalidArgumentException(
            '--cluster-zone', '--cluster-zone must be specified.')
      cluster = msgs.Cluster(
          serveNodes=arguments.ProcessInstanceTypeAndNodes(args),
          # nodeScalingFactor is not supported in deprecated workflow
          defaultStorageType=storage_type,
          nodeScalingFactor=msgs.Cluster.NodeScalingFactorValueValuesEnum(
              msgs.Cluster.NodeScalingFactorValueValuesEnum.NODE_SCALING_FACTOR_1X
          ),
          # TODO(b/36049938): switch location to resource
          # when b/29566669 is fixed on API
          location=util.LocationUrl(args.cluster_zone))
      return {args.cluster: cluster}
    else:
      raise exceptions.InvalidArgumentException(
          '--cluster --cluster-config',
          'Use --cluster-config to specify cluster(s).',
      )

  def _ValidateClusterConfigArgs(self, cluster_config):
    """Validates arguments in cluster-config as a repeated dict."""
    # Validate cluster-config of each cluster.
    for cluster_dict in cluster_config:
      if ('autoscaling-min-nodes' in cluster_dict or
          'autoscaling-max-nodes' in cluster_dict or
          'autoscaling-cpu-target' in cluster_dict or
          'autoscaling-storage-target' in cluster_dict):
        # nodes and autoscaling args are mutual exclusive.
        if 'nodes' in cluster_dict:
          raise exceptions.InvalidArgumentException(
              '--autoscaling-min-nodes --autoscaling-max-nodes '
              '--autoscaling-cpu-target --autoscaling-storage-target',
              'At most one of nodes | autoscaling-min-nodes '
              'autoscaling-max-nodes autoscaling-cpu-target '
              'autoscaling-storage-target may be specified in --cluster-config')
        # To enable autoscaling, all related args must be set.
        if ('autoscaling-min-nodes' not in cluster_dict or
            'autoscaling-max-nodes' not in cluster_dict or
            'autoscaling-cpu-target' not in cluster_dict):
          raise exceptions.InvalidArgumentException(
              '--autoscaling-min-nodes --autoscaling-max-nodes '
              '--autoscaling-cpu-target', 'All of --autoscaling-min-nodes '
              '--autoscaling-max-nodes --autoscaling-cpu-target must be set to '
              'enable Autoscaling.')

  @classmethod
  def _Tags(cls, args, tags_arg_name='tags'):
    """Get the tags from command arguments.

    Args:
      args: the argparse namespace from Run().
      tags_arg_name: the name of the tags argument.

    Returns:
      A dict mapping from tag key to tag value.
    """

    tags = getattr(args, tags_arg_name)
    if not tags:
      return None

    tags_message = util.GetAdminMessages().Instance.TagsValue
    # Sorted for test stability.
    return tags_message(
        additionalProperties=[
            tags_message.AdditionalProperty(key=key, value=value)
            for key, value in sorted(tags.items())
        ]
    )


@base.ReleaseTracks(base.ReleaseTrack.ALPHA)
class CreateInstanceAlpha(CreateInstance):
  """Create a new Bigtable instance."""

  @classmethod
  def Args(cls, parser):
    super(CreateInstanceAlpha, cls).Args(parser)



# -*- coding: utf-8 -*- #
# Copyright 2020 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""services api-keys update command."""

from __future__ import annotations

import argparse
from collections.abc import Callable, Sequence
import dataclasses
import itertools
import types
from typing import Any, TypeVar

from googlecloudsdk.api_lib.services import apikeys
from googlecloudsdk.api_lib.services import services_util
from googlecloudsdk.calliope import base
from googlecloudsdk.calliope import exceptions
from googlecloudsdk.command_lib.services import common_flags
from googlecloudsdk.core import log

OP_BASE_CMD = 'gcloud services operations '
OP_WAIT_CMD = OP_BASE_CMD + 'wait {0}'
DETAILED_HELP = {
    'EXAMPLES': (
        """
        To remove all restrictions of the key:

          $ {command} projects/myproject/locations/global/keys/my-key-id --clear-restrictions

        To update display name and set allowed ips as server key restrictions:

          $ {command} projects/myproject/locations/global/keys/my-key-id --display-name="test name" --allowed-ips=2620:15c:2c4:203:2776:1f90:6b3b:217,104.133.8.78

        To update annotations:

          $ {command} projects/myproject/locations/global/keys/my-key-id --annotations=foo=bar,abc=def

        To update key's allowed referrers restriction:

          $ {command} projects/myproject/locations/global/keys/my-key-id --allowed-referrers="https://www.example.com/*,http://sub.example.com/*"

        To update key's allowed ios app bundle ids:

          $ {command} projects/myproject/locations/global/keys/my-key-id --allowed-bundle-ids=my.app

        To update key's allowed android application:

          $ {command} projects/myproject/locations/global/keys/my-key-id --allowed-application=sha1_fingerprint=foo1,package_name=bar1 --allowed-application=sha1_fingerprint=foo2,package_name=bar2

        To update keys' allowed api target with multiple services:

          $ {command} projects/myproject/locations/global/keys/my-key-id --api-target=service=bar.service.com --api-target=service=foo.service.com

        To update keys' allowed api target with service and method:

          $ {command} projects/myproject/locations/global/keys/my-key-id  --flags-file=my-flags.yaml

          The content of 'my-flags.yaml' is as following:

          ```
            - --api-target:
                service: "foo.service.com"
            - --api-target:
                service: "bar.service.com"
                methods:
                - "foomethod"
                - "barmethod"
            ```
        """
    )
}


@base.UniverseCompatible
@base.ReleaseTracks(base.ReleaseTrack.GA, base.ReleaseTrack.BETA)
class Update(base.UpdateCommand):
  """Update an API key's metadata."""

  @staticmethod
  def Args(parser):
    common_flags.key_flag(parser=parser, suffix='to update')
    common_flags.display_name_flag(parser=parser, suffix='to update')
    common_flags.add_key_update_args(parser)
    base.ASYNC_FLAG.AddToParser(parser)

  def Run(self, args: argparse.Namespace) -> Any:
    """Run command.

    Args:
      args: an argparse namespace. All the arguments that were provided to this
        command invocation.

    Returns:
      None
    """

    client = apikeys.GetClientInstance()
    messages = client.MESSAGES_MODULE

    key_ref = args.CONCEPTS.key.Parse()
    update_mask = []
    key_proto = messages.V2Key(
        name=key_ref.RelativeName(), restrictions=messages.V2Restrictions()
    )
    if args.IsSpecified('annotations'):
      update_mask.append('annotations')
      key_proto.annotations = apikeys.GetAnnotations(args, messages)
    if args.IsSpecified('display_name'):
      update_mask.append('display_name')
      key_proto.displayName = args.display_name
    if args.IsSpecified('clear_annotations'):
      update_mask.append('annotations')
    if args.IsSpecified('clear_restrictions'):
      update_mask.append('restrictions')
    else:
      if args.IsSpecified('allowed_referrers'):
        update_mask.append('restrictions.browser_key_restrictions')
        key_proto.restrictions.browserKeyRestrictions = (
            messages.V2BrowserKeyRestrictions(
                allowedReferrers=args.allowed_referrers
            )
        )
      elif args.IsSpecified('allowed_ips'):
        update_mask.append('restrictions.server_key_restrictions')
        key_proto.restrictions.serverKeyRestrictions = (
            messages.V2ServerKeyRestrictions(allowedIps=args.allowed_ips)
        )
      elif args.IsSpecified('allowed_bundle_ids'):
        update_mask.append('restrictions.ios_key_restrictions')
        key_proto.restrictions.iosKeyRestrictions = (
            messages.V2IosKeyRestrictions(
                allowedBundleIds=args.allowed_bundle_ids
            )
        )
      elif args.IsSpecified('allowed_application'):
        update_mask.append('restrictions.android_key_restrictions')
        key_proto.restrictions.androidKeyRestrictions = (
            messages.V2AndroidKeyRestrictions(
                allowedApplications=apikeys.GetAllowedAndroidApplications(
                    args, messages
                )
            )
        )
      if args.IsSpecified('api_target'):
        update_mask.append('restrictions.api_targets')
        key_proto.restrictions.apiTargets = apikeys.GetApiTargets(
            args, messages
        )
    request = messages.ApikeysProjectsLocationsKeysPatchRequest(
        name=key_ref.RelativeName(),
        updateMask=','.join(update_mask),
        v2Key=key_proto,
    )
    op = client.projects_locations_keys.Patch(request)
    if not op.done:
      if args.async_:
        cmd = OP_WAIT_CMD.format(op.name)
        log.status.Print(
            f'Asynchronous operation is in progress... Use the following '
            f'command to wait for its completion:\n {cmd}'
        )
        return op
      op = services_util.WaitOperation(op.name, apikeys.GetOperation)
    services_util.PrintOperationWithResponse(op)
    return op

  detailed_help = DETAILED_HELP


@base.UniverseCompatible
@base.ReleaseTracks(base.ReleaseTrack.ALPHA)
class AlphaUpdate(Update):
  """A surface for updating API keys, including append support."""

  @classmethod
  def Args(cls, parser):
    """Add arguments to the parser.

    Args:
      parser: The argparse.ArgumentParser to which the arguments are added.
    """
    Update.Args(parser)
    parser.add_argument(
        '--append',
        action='store_true',
        help=(
            'If specified, merge the new restrictions with the current '
            'restrictions of the key instead of replacing them.'
        ),
    )

  def Run(self, args: argparse.Namespace) -> Any:
    """Run the update command for Alpha release track.

    This method extends the base Update.Run to support the '--append' flag,
    which allows merging new restrictions with existing ones instead of
    replacing them.

    Args:
      args: An argparse namespace. All the arguments that were provided to this
        command invocation.

    Returns:
      The LRO object.

    Raises:
      googlecloudsdk.calliope.exceptions.InvalidArgumentException:
        If the --clear-restrictions flag is specified, if the updated key
        lacks API target restrictions, or if attempting to append a different
        client restriction type than already exists on the key.
    """
    client = apikeys.GetClientInstance(self.ReleaseTrack())
    messages = client.MESSAGES_MODULE
    key_ref = args.CONCEPTS.key.Parse()
    request = messages.ApikeysProjectsLocationsKeysGetRequest(
        name=key_ref.RelativeName()
    )
    current_key = client.projects_locations_keys.Get(request)

    current_has_api_targets = (
        current_key.restrictions is not None
        and current_key.restrictions.apiTargets
    )
    is_clearing = args.IsSpecified('clear_restrictions')
    is_specifying_api_target = args.IsSpecified('api_target')

    if is_clearing:
      raise exceptions.InvalidArgumentException(
          '--clear-restrictions',
          'The --clear-restrictions flag is not supported on the ALPHA track.',
      )

    if not (current_has_api_targets or is_specifying_api_target):
      raise exceptions.InvalidArgumentException(
          '--api-target',
          'API keys must have API target restrictions. Please specify '
          '`--api-target` to restrict this key.',
      )

    if not args.append:
      return super(AlphaUpdate, self).Run(args)

    update_mask = []
    key_proto = messages.V2Key(
        name=key_ref.RelativeName(), restrictions=messages.V2Restrictions()
    )

    if args.IsSpecified('annotations'):
      update_mask.append('annotations')
      key_proto.annotations = apikeys.GetAnnotations(args, messages)
    if args.IsSpecified('display_name'):
      update_mask.append('display_name')
      key_proto.displayName = args.display_name
    if args.IsSpecified('clear_annotations'):
      update_mask.append('annotations')

    # Copy etag from current key to ensure the update is based on the
    # latest version, preventing race conditions.
    if current_key.etag:
      key_proto.etag = current_key.etag

    current_restrictions = getattr(current_key, 'restrictions', None)
    current_client_type = _GetClientRestrictionType(current_restrictions)

    new_client_type = None
    if args.IsSpecified('allowed_referrers'):
      new_client_type = 'browserKeyRestrictions'
    elif args.IsSpecified('allowed_ips'):
      new_client_type = 'serverKeyRestrictions'
    elif args.IsSpecified('allowed_bundle_ids'):
      new_client_type = 'iosKeyRestrictions'
    elif args.IsSpecified('allowed_application'):
      new_client_type = 'androidKeyRestrictions'

    if (
        new_client_type
        and current_client_type
        and current_client_type != new_client_type
    ):
      new_friendly = _FRIENDLY_NAME_BY_TYPE[new_client_type]
      current_friendly = _FRIENDLY_NAME_BY_TYPE[current_client_type]
      raise exceptions.InvalidArgumentException(
          f'--{new_friendly}',
          f'Cannot append {new_friendly} restriction because the key already '
          f'has {current_friendly} restriction.',
      )

    if new_client_type:
      field_name, restriction_msg, mask_path = _GetMergedApplicationRestriction(
          current_restrictions, new_client_type, args, messages
      )
      setattr(key_proto.restrictions, field_name, restriction_msg)
      update_mask.append(mask_path)

    if args.IsSpecified('api_target'):
      current_api_targets = (
          getattr(current_restrictions, 'apiTargets', [])
          if current_restrictions
          else []
      )
      if current_api_targets is None:
        current_api_targets = []
      new_api_targets = apikeys.GetApiTargets(args, messages) or []
      merged_targets = _MergeApiTargets(
          current_api_targets, new_api_targets, messages
      )

      key_proto.restrictions.apiTargets = merged_targets
      update_mask.append('restrictions.api_targets')

    # Construct and send the patch request.
    request = messages.ApikeysProjectsLocationsKeysPatchRequest(
        name=key_ref.RelativeName(),
        updateMask=','.join(update_mask),
        v2Key=key_proto,
    )
    op = client.projects_locations_keys.Patch(request)
    if not op.done:
      if args.async_:
        cmd = OP_WAIT_CMD.format(op.name)
        log.status.Print(
            f'Asynchronous operation is in progress... Use the following '
            f'command to wait for its completion:\n {cmd}'
        )
        return op
      op = services_util.WaitOperation(op.name, apikeys.GetOperation)
    services_util.PrintOperationWithResponse(op)
    return op


@dataclasses.dataclass
class _ApiTargetEntry:
  """A structured record representing a parsed API target service.

  Attributes:
    service: The display service name of the target.
    methods: The list of methods allowed for this target.
    all_allowed: True if all methods are allowed for this service.
  """
  service: str
  methods: list[str]
  all_allowed: bool


def _GetAndroidAppKey(app: Any) -> tuple[str, str]:
  """Returns a package/fingerprint tuple representing an Android app."""
  pkg = getattr(app, 'packageName', '') or ''
  fingerprint = _NormalizeSha1(getattr(app, 'sha1Fingerprint', ''))
  return (pkg.strip(), fingerprint)


def _GetApiTargetMethods(target: Any) -> list[str]:
  """Returns a list of methods allowed for an API target."""
  return list(getattr(target, 'methods', None) or [])


# TODO(b/517632542): This method assumes client restrictions are mutually
# exclusive. This will need to be updated when API V3 is integrated,
# which supports multiple restriction types.
def _GetClientRestrictionType(restrictions: Any | None) -> str | None:
  """Returns the active application restriction type string from current restrictions.

  Args:
    restrictions: The key's restrictions protobuf message.

  Returns:
    The active restriction field name (str) or None if not set.
  """
  if restrictions is None:
    return None
  browser_key_restrictions = getattr(
      restrictions, 'browserKeyRestrictions', None
  )
  if (
      browser_key_restrictions is not None
      and getattr(browser_key_restrictions, 'allowedReferrers', None)
      is not None
  ):
    return 'browserKeyRestrictions'
  server_key_restrictions = getattr(restrictions, 'serverKeyRestrictions', None)
  if (
      server_key_restrictions is not None
      and getattr(server_key_restrictions, 'allowedIps', None) is not None
  ):
    return 'serverKeyRestrictions'
  ios_key_restrictions = getattr(restrictions, 'iosKeyRestrictions', None)
  if (
      ios_key_restrictions is not None
      and getattr(ios_key_restrictions, 'allowedBundleIds', None) is not None
  ):
    return 'iosKeyRestrictions'
  android_key_restrictions = getattr(
      restrictions, 'androidKeyRestrictions', None
  )
  if (
      android_key_restrictions is not None
      and getattr(android_key_restrictions, 'allowedApplications', None)
      is not None
  ):
    return 'androidKeyRestrictions'
  return None


def _NormalizeSha1(fingerprint: str | None) -> str:
  """Normalizes a SHA-1 fingerprint by stripping colons and spaces.

  Args:
    fingerprint: The raw fingerprint string to normalize.

  Returns:
    The normalized lowercase fingerprint (str).
  """
  if fingerprint is None:
    return ''
  return fingerprint.replace(':', '').replace(' ', '').lower()


_T = TypeVar('_T')


def _UnionPreservingOrder(
    current_list: Sequence[_T] | None,
    new_list: Sequence[_T] | None,
    key_fn: Callable[[_T], Any] = lambda x: x,
) -> list[_T]:
  """Unions two lists while preserving original insertion order.

  Args:
    current_list: The existing list.
    new_list: The new list to append.
    key_fn: A function to derive the comparison key for each item.

  Returns:
    A deduplicated union list preserving original insertion order.
  """
  seen = set()
  result = []
  for item in itertools.chain(current_list or [], new_list or []):
    if item is not None:
      key = key_fn(item)
      if key not in seen:
        seen.add(key)
        result.append(item)
  return result


def _GetMergedApplicationRestriction(
    current_restrictions: Any | None,
    client_type: str,
    args: argparse.Namespace,
    messages: Any,
) -> tuple[str, Any, str]:
  """Merges current and new application restrictions.

  Args:
    current_restrictions: The current V2Restrictions message.
    client_type: The client restriction type string.
    args: The argparse namespace.
    messages: The API client messages module.

  Returns:
    A tuple of (field_name, restriction_message, update_mask_path).
  """
  if client_type == 'browserKeyRestrictions':
    current_browser = (
        getattr(current_restrictions, 'browserKeyRestrictions', None)
        if current_restrictions is not None
        else None
    )
    current_list = (
        getattr(current_browser, 'allowedReferrers', [])
        if current_browser is not None
        else []
    )
    if current_list is None:
      current_list = []
    new_list = args.allowed_referrers or []
    merged_list = _UnionPreservingOrder(current_list, new_list)
    return (
        'browserKeyRestrictions',
        messages.V2BrowserKeyRestrictions(allowedReferrers=merged_list),
        'restrictions.browser_key_restrictions',
    )

  if client_type == 'serverKeyRestrictions':
    current_server = (
        getattr(current_restrictions, 'serverKeyRestrictions', None)
        if current_restrictions is not None
        else None
    )
    current_list = (
        getattr(current_server, 'allowedIps', [])
        if current_server is not None
        else []
    )
    if current_list is None:
      current_list = []
    new_list = args.allowed_ips or []
    merged_list = _UnionPreservingOrder(current_list, new_list)
    return (
        'serverKeyRestrictions',
        messages.V2ServerKeyRestrictions(allowedIps=merged_list),
        'restrictions.server_key_restrictions',
    )

  if client_type == 'iosKeyRestrictions':
    current_ios = (
        getattr(current_restrictions, 'iosKeyRestrictions', None)
        if current_restrictions is not None
        else None
    )
    current_list = (
        getattr(current_ios, 'allowedBundleIds', [])
        if current_ios is not None
        else []
    )
    if current_list is None:
      current_list = []
    new_list = args.allowed_bundle_ids or []
    merged_list = _UnionPreservingOrder(current_list, new_list)
    return (
        'iosKeyRestrictions',
        messages.V2IosKeyRestrictions(allowedBundleIds=merged_list),
        'restrictions.ios_key_restrictions',
    )

  if client_type == 'androidKeyRestrictions':
    current_android = (
        getattr(current_restrictions, 'androidKeyRestrictions', None)
        if current_restrictions is not None
        else None
    )
    current_list = (
        getattr(current_android, 'allowedApplications', [])
        if current_android is not None
        else []
    )
    if current_list is None:
      current_list = []
    new_list = apikeys.GetAllowedAndroidApplications(args, messages) or []
    merged_list = _UnionPreservingOrder(
        current_list, new_list, key_fn=_GetAndroidAppKey
    )
    return (
        'androidKeyRestrictions',
        messages.V2AndroidKeyRestrictions(allowedApplications=merged_list),
        'restrictions.android_key_restrictions',
    )

  raise ValueError(f'Unknown client type: {client_type}')


def _MergeApiTargets(
    current_targets: Sequence[Any] | None,
    new_targets: Sequence[Any] | None,
    messages: Any,
) -> list[Any]:
  """Merges current and new API target configurations by service.

  Args:
    current_targets: The list of current V2ApiTarget messages.
    new_targets: The list of new V2ApiTarget messages.
    messages: The API client messages module.

  Returns:
    A list of merged V2ApiTarget messages.
  """
  target_map = {}
  ordered_services: list[str] = []

  for target in current_targets or []:
    if target is None or getattr(target, 'service', None) is None:
      continue
    srv = target.service
    srv_lower = srv.lower()
    if srv_lower not in target_map:
      ordered_services.append(srv_lower)
    target_map[srv_lower] = _ApiTargetEntry(
        service=srv,
        methods=_GetApiTargetMethods(target),
        all_allowed=not getattr(target, 'methods', None),
    )

  for target in new_targets or []:
    if target is None or getattr(target, 'service', None) is None:
      continue
    srv = target.service
    srv_lower = srv.lower()
    new_methods = _GetApiTargetMethods(target)
    new_all_allowed = not getattr(target, 'methods', None)

    entry = target_map.get(srv_lower)
    if entry is not None:
      if entry.all_allowed or new_all_allowed:
        entry.all_allowed = True
        entry.methods = []
      else:
        entry.methods = _UnionPreservingOrder(
            entry.methods, new_methods, key_fn=lambda x: x.lower()
        )
    else:
      ordered_services.append(srv_lower)
      target_map[srv_lower] = _ApiTargetEntry(
          service=srv,
          methods=new_methods,
          all_allowed=new_all_allowed,
      )

  return [
      messages.V2ApiTarget(
          service=target_map[srv_lower].service,
          methods=(
              target_map[srv_lower].methods
              if not target_map[srv_lower].all_allowed
              else []
          ),
      )
      for srv_lower in ordered_services
  ]


_FRIENDLY_NAME_BY_TYPE = types.MappingProxyType({
    'browserKeyRestrictions': 'allowed-referrers',
    'serverKeyRestrictions': 'allowed-ips',
    'iosKeyRestrictions': 'allowed-bundle-ids',
    'androidKeyRestrictions': 'allowed-application',
})



# -*- coding: utf-8 -*- #
# Copyright 2017 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The 'gcloud firebase test android run' command."""


from googlecloudsdk.api_lib.firebase.test import arg_util
from googlecloudsdk.api_lib.firebase.test import ctrl_c_handler
from googlecloudsdk.api_lib.firebase.test import exit_code
from googlecloudsdk.api_lib.firebase.test import history_picker
from googlecloudsdk.api_lib.firebase.test import matrix_ops
from googlecloudsdk.api_lib.firebase.test import results_bucket
from googlecloudsdk.api_lib.firebase.test import results_summary
from googlecloudsdk.api_lib.firebase.test import tool_results
from googlecloudsdk.api_lib.firebase.test import util
from googlecloudsdk.api_lib.firebase.test.android import arg_manager
from googlecloudsdk.api_lib.firebase.test.android import matrix_creator
from googlecloudsdk.calliope import base
from googlecloudsdk.core import log

_APK_MIME_TYPE = 'application/vnd.android.package-archive'


class _BaseRun(object):
  """Invoke a test in Firebase Test Lab for Android and view test results."""

  detailed_help = {
      'DESCRIPTION': """
          *{command}* invokes and monitors tests in Firebase Test Lab for
          Android.

          Three main types of Android tests are currently supported:
          - *robo*: runs a smart, automated exploration of the activities in
            your Android app which records any installation failures or crashes
            and builds an activity map with associated screenshots and video.
          - *instrumentation*: runs automated unit or integration tests written
            using a testing framework. Firebase Test Lab for Android currently
            supports the Espresso and UI Automator 2.0 testing
            frameworks.
          - *game-loop*: executes a special intent built into the game app (a
            "demo mode") that simulates the actions of a real player. This test
            type can include multiple game loops (also called "scenarios"),
            which can be logically organized using scenario labels so that you
            can run related loops together. Refer to
            https://firebase.google.com/docs/test-lab/android/game-loop for
            more information about how to build and run Game Loop tests.

          The type of test to run can be specified with the *--type* flag,
          although the type can often be inferred from other flags.
          Specifically, if the *--test* flag is present, the test *--type*
          defaults to `instrumentation`. If *--test* is not present, then
          *--type* defaults to `robo`.

          All arguments for *{command}* may be specified on the command line
          and/or within an argument file. Run *$ gcloud topic arg-files* for
          more information about argument files.
          """,
      'EXAMPLES': """
          To invoke a robo test lasting 100 seconds against the default device
          environment, run:

            $ {command} --app=APP_APK --timeout=100s

          When specifying devices to test against, the preferred method is to
          use the --device flag. For example, to invoke a robo test against a
          virtual, generic MDPI Nexus device in landscape orientation, run:

            $ {command} --app=APP_APK --device=model=NexusLowRes,orientation=landscape

          To invoke an instrumentation test against a physical Nexus 6 device
          (MODEL_ID: shamu) which is running Android API level 21 in French, run:

            $ {command} --app=APP_APK --test=TEST_APK --device=model=shamu,version=21,locale=fr

          To test against multiple devices, specify --device more than once:

            $ {command} --app=APP_APK --test=TEST_APK --device=model=Nexus4,version=19 --device=model=Nexus4,version=21 --device=model=NexusLowRes,version=25

          To invoke a robo test on an Android App Bundle, pass the .aab file
          using the --app flag.

            $ {command} --app=bundle.aab

          You may also use the legacy dimension flags (deprecated) to specify
          which devices to use. Firebase Test Lab will run tests against every
          possible combination of the listed device dimensions. Note that some
          combinations of device models and OS versions may not be valid or
          available in Test Lab. Any unsupported combinations of dimensions in
          the test matrix will be skipped.

          For example, to execute a series of 5-minute robo tests against a very
          comprehensive matrix of virtual and physical devices, OS versions,
          locales and orientations, run:

            $ {command} --app=APP_APK --timeout=5m --device-ids=shamu,NexusLowRes,Nexus5,g3,zeroflte --os-version-ids=19,21,22,23,24,25 --locales=en_GB,es,fr,ru,zh --orientations=portrait,landscape

          The above command will generate a test matrix with a total of 300 test
          executions, but only the subset of executions with valid dimension
          combinations will actually run your tests.

          To help you identify and locate your test matrix in the Firebase
          console, run:

            $ {command} --app=APP_APK --client-details=matrixLabel="Example matrix label"

          Controlling Results Storage

          By default, Firebase Test Lab stores detailed test results for a
          limited time in a Google Cloud Storage bucket provided for you at
          no charge. Note: This requires the principal executing the test to
          have "roles/editor" role for the project. If either you do not have
          that role, you wish to use a storage bucket that you control, or you
          need to retain detailed test results for a longer period, use the
          *--results-bucket* option. See
          https://firebase.google.com/docs/test-lab/analyzing-results#detailed
          for more information.

          Detailed test result files are prefixed by default with a timestamp
          and a random character string. If you require a predictable path
          where detailed test results are stored within the results bucket
          (say, if you have a Continuous Integration system which does custom
          post-processing of test result artifacts), use the *--results-dir*
          option. _Note that each test invocation *must* have a unique storage
          location, so never reuse the same value for *--results-dir* between
          different test runs_. Possible strategies could include using a UUID
          or sequence number for *--results-dir*.

          For example, to run a robo test using a specific Google Cloud Storage
          location to hold the raw test results, run:

            $ {command} --app=APP_APK --results-bucket=gs://my-bucket --results-dir=my/test/results/<unique-value>

          To run an instrumentation test and specify a custom name under which
          the history of your tests will be collected and displayed in the
          Firebase console, run:

            $ {command} --app=APP_APK --test=TEST_APK --results-history-name='Excelsior App Test History'

          Argument Files

          All test arguments for a given test may alternatively be stored in an
          argument group within a YAML-formatted argument file. The _ARG_FILE_
          may contain one or more named argument groups, and argument groups may
          be combined using the `include:` attribute (Run *$ gcloud topic
          arg-files* for more information). The ARG_FILE can easily be shared
          with colleagues or placed under source control to ensure consistent
          test executions.

          To run a test using arguments loaded from an ARG_FILE named
          *excelsior_args*, which contains an argument group named *robo-args:*,
          use the following syntax:

            $ {command} path/to/excelsior_args:robo-args
          """,
  }

  def Run(self, args):
    """Run the 'gcloud firebase test run' command to invoke a test in Test Lab.

    Args:
      args: an argparse namespace. All the arguments that were provided to this
        command invocation (i.e. group and command arguments combined).

    Returns:
      One of:
        - a list of TestOutcome tuples (if ToolResults are available).
        - a URL string pointing to the user's results in ToolResults or GCS.
    """
    if args.async_ and not args.IsSpecified('format'):
      args.format = """
          value(format(
            'Final test results will be available at [ {0} ].', [])
          )
      """
    log.status.Print('\nHave questions, feedback, or issues? Get support by '
                     'visiting:\n  https://firebase.google.com/support/\n')

    arg_manager.AndroidArgsManager().Prepare(args)

    project = util.GetProject()
    tr_client = self.context['toolresults_client']
    tr_messages = self.context['toolresults_messages']
    storage_client = self.context['storage_client']

    bucket_ops = results_bucket.ResultsBucketOps(project, args.results_bucket,
                                                 args.results_dir, tr_client,
                                                 tr_messages, storage_client)
    bucket_ops.UploadFileToGcs(args.app, _APK_MIME_TYPE)
    if args.test:
      bucket_ops.UploadFileToGcs(args.test, _APK_MIME_TYPE)
    for obb_file in (args.obb_files or []):
      bucket_ops.UploadFileToGcs(obb_file, 'application/octet-stream')
    if getattr(args, 'robo_script', None):
      bucket_ops.UploadFileToGcs(args.robo_script, 'application/json')
    additional_apks = getattr(args, 'additional_apks', None) or []
    for additional_apk in additional_apks:
      bucket_ops.UploadFileToGcs(additional_apk, _APK_MIME_TYPE)
    other_files = getattr(args, 'other_files', None) or {}
    for device_path, file_to_upload in other_files.items():
      bucket_ops.UploadFileToGcs(
          file_to_upload,
          None,
          destination_object=util.GetRelativeDevicePath(device_path))
    bucket_ops.LogGcsResultsUrl()

    tr_history_picker = history_picker.ToolResultsHistoryPicker(
        project, tr_client, tr_messages)
    history_name = PickHistoryName(args)
    history_id = tr_history_picker.GetToolResultsHistoryId(history_name)

    matrix = matrix_creator.CreateMatrix(
        args,
        self.context,
        history_id,
        bucket_ops.gcs_results_root,
        str(self.ReleaseTrack()),
    )
    monitor = matrix_ops.MatrixMonitor(matrix.testMatrixId, args.type,
                                       self.context)

    with ctrl_c_handler.CancellableTestSection(monitor):
      tr_ids = tool_results.GetToolResultsIds(matrix, monitor)
      matrix = monitor.GetTestMatrixStatus()
      supported_executions = monitor.HandleUnsupportedExecutions(matrix)

      url = tool_results.CreateToolResultsUiUrl(project, tr_ids)
      log.status.Print('')
      if args.async_:
        return url
      log.status.Print('Test results will be streamed to [ {0} ].'.format(url))

      # If we have exactly one testExecution, show detailed progress info.
      if len(supported_executions) == 1 and args.num_flaky_test_attempts == 0:
        monitor.MonitorTestExecutionProgress(supported_executions[0].id)
      else:
        monitor.MonitorTestMatrixProgress()

    log.status.Print('\nMore details are available at [ {0} ].'.format(url))
    # Fetch the per-dimension test outcomes list, and also the "rolled-up"
    # matrix outcome from the Tool Results service.
    summary_fetcher = results_summary.ToolResultsSummaryFetcher(
        project, tr_client, tr_messages, tr_ids, matrix.testMatrixId)
    self.exit_code = exit_code.ExitCodeFromRollupOutcome(
        summary_fetcher.FetchMatrixRollupOutcome(),
        tr_messages.Outcome.SummaryValueValuesEnum)
    return summary_fetcher.CreateMatrixOutcomeSummaryUsingEnvironments()


@base.ReleaseTracks(base.ReleaseTrack.GA)
class RunGA(_BaseRun, base.ListCommand):
  """Invoke a test in Firebase Test Lab for Android and view test results."""

  @staticmethod
  def Args(parser):
    arg_util.AddCommonTestRunArgs(parser)
    arg_util.AddMatrixArgs(parser)
    arg_util.AddAndroidTestArgs(parser)
    arg_util.AddGaArgs(parser)
    base.URI_FLAG.RemoveFromParser(parser)
    parser.display_info.AddFormat(util.OUTCOMES_FORMAT)


@base.ReleaseTracks(base.ReleaseTrack.ALPHA, base.ReleaseTrack.BETA)
class RunBeta(_BaseRun, base.ListCommand):
  """Invoke a test in Firebase Test Lab for Android and view test results."""

  @staticmethod
  def Args(parser):
    arg_util.AddCommonTestRunArgs(parser)
    arg_util.AddMatrixArgs(parser)
    arg_util.AddAndroidTestArgs(parser)
    arg_util.AddAndroidBetaArgs(parser)
    arg_util.AddBetaArgs(parser)
    base.URI_FLAG.RemoveFromParser(parser)
    parser.display_info.AddFormat(util.OUTCOMES_FORMAT)


def PickHistoryName(args):
  """Returns the results history name to use to look up a history ID.

  The history ID corresponds to a history name. If the user provides their
  own history name, we use that to look up the history ID; If not, but the user
  provides an app-package name, we use the app-package name with ' (gcloud)'
  appended as the history name. Otherwise, we punt and let the Testing service
  determine the appropriate history ID to publish to.

  Args:
    args: an argparse namespace. All the arguments that were provided to the
      command invocation (i.e. group and command arguments combined).

  Returns:
    Either a string containing a history name derived from user-supplied data,
    or None if we lack the required information.
  """
  if args.results_history_name:
    return args.results_history_name
  if args.app_package:
    return args.app_package + ' (gcloud)'
  return None



# -*- coding: utf-8 -*- #
# Copyright 2016 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Command for spanner instances create."""

import textwrap

from googlecloudsdk.api_lib.spanner import instance_operations
from googlecloudsdk.api_lib.spanner import instances
from googlecloudsdk.calliope import base
from googlecloudsdk.command_lib.spanner import flags
from googlecloudsdk.command_lib.spanner import resource_args
from googlecloudsdk.command_lib.util.args import labels_util


@base.UniverseCompatible
@base.ReleaseTracks(base.ReleaseTrack.GA)
class Create(base.CreateCommand):
  """Create a Cloud Spanner instance."""

  detailed_help = {
      'EXAMPLES':
          textwrap.dedent("""\
        To create a Cloud Spanner instance, run:

          $ {command} my-instance-id --config=regional-us-east1 --description=my-instance-display-name --nodes=3
        """),
  }

  @staticmethod
  def Args(parser):
    """Args is called by calliope to gather arguments for this command.

    Please add arguments in alphabetical order except for no- or a clear-
    pair for that argument which can follow the argument itself.
    Args:
      parser: An argparse parser that you can use to add arguments that go
          on the command line after this command. Positional arguments are
          allowed.
    """
    flags.Instance().AddToParser(parser)
    flags.Config().AddToParser(parser)
    flags.Description().AddToParser(parser)
    flags.Edition(
        choices={
            'STANDARD': 'Standard edition',
            'ENTERPRISE': 'Enterprise edition',
            'ENTERPRISE_PLUS': 'Enterprise Plus edition',
            'EDITION_UNSPECIFIED': (
                "Spanner's legacy pricing model. For more information, see the"
                ' [Spanner editions overview]'
                '(https://cloud.google.com/spanner/docs/editions-overview)'
            ),
        },
    ).AddToParser(parser)
    flags.DefaultBackupScheduleType(
        choices={
            'DEFAULT_BACKUP_SCHEDULE_TYPE_UNSPECIFIED': 'Not specified.',
            'NONE': (
                'No default backup schedule is created automatically when a new'
                ' database is created in an instance.'
            ),
            'AUTOMATIC': (
                'A default backup schedule is created automatically when a new'
                ' database is created in an instance. You can edit or delete'
                " the default backup schedule once it's created. The default"
                ' backup schedule creates a full backup every 24 hours. These'
                ' full backups are retained for 7 days.'
            ),
        },
    ).AddToParser(parser)
    resource_args.AddExpireBehaviorArg(parser)
    resource_args.AddInstanceTypeArg(parser)
    flags.AddCapacityArgsForInstance(
        require_all_autoscaling_args=True,
        parser=parser,
        add_asymmetric_option_flag=True,
        autoscaling_cpu_target_group=True,
        add_asymmetric_total_cpu_target_flag=True,
        add_asymmetric_disable_autoscaling_flags=True,
        add_disable_downscaling_flag=True,
    )
    base.ASYNC_FLAG.AddToParser(parser)
    parser.display_info.AddCacheUpdater(flags.InstanceCompleter)
    flags.AddTags(parser)
    labels_util.AddCreateLabelsFlags(parser)

  def Run(self, args):
    """This is what gets called when the user runs this command.

    Args:
      args: an argparse namespace. All the arguments that were provided to this
        command invocation.

    Returns:
      Some value that we want to have printed later.
    """
    instance_type = resource_args.GetInstanceType(args)
    expire_behavior = resource_args.GetExpireBehavior(args)

    op = instances.Create(
        instance=args.instance,
        config=args.config,
        description=args.description,
        nodes=args.nodes,
        processing_units=args.processing_units,
        autoscaling_min_nodes=args.autoscaling_min_nodes,
        autoscaling_max_nodes=args.autoscaling_max_nodes,
        autoscaling_min_processing_units=args.autoscaling_min_processing_units,
        autoscaling_max_processing_units=args.autoscaling_max_processing_units,
        autoscaling_high_priority_cpu_target=args.autoscaling_high_priority_cpu_target,
        autoscaling_total_cpu_target=args.autoscaling_total_cpu_target,
        autoscaling_storage_target=args.autoscaling_storage_target,
        asymmetric_autoscaling_options=args.asymmetric_autoscaling_option,
        disable_downscaling=args.disable_downscaling,
        instance_type=instance_type,
        expire_behavior=expire_behavior,
        edition=args.edition,
        default_backup_schedule_type=args.default_backup_schedule_type,
        tags=args.tags,
        labels=args.labels,
    )
    if args.async_:
      return op
    instance_operations.Await(op, 'Creating instance')


@base.ReleaseTracks(base.ReleaseTrack.BETA)
@base.UniverseCompatible
class BetaCreate(base.CreateCommand):
  """Create a Cloud Spanner instance."""

  detailed_help = {
      'EXAMPLES':
          textwrap.dedent("""\
        To create a Cloud Spanner instance, run:

          $ {command} my-instance-id --config=regional-us-east1 --description=my-instance-display-name --nodes=3
        """),
  }

  @staticmethod
  def Args(parser):
    """Args is called by calliope to gather arguments for this command.

    Please add arguments in alphabetical order except for no- or a clear-
    pair for that argument which can follow the argument itself.
    Args:
      parser: An argparse parser that you can use to add arguments that go
          on the command line after this command. Positional arguments are
          allowed.
    """
    flags.Instance().AddToParser(parser)
    flags.Config().AddToParser(parser)
    flags.Description().AddToParser(parser)
    flags.Edition(
        choices={
            'STANDARD': 'Standard edition',
            'ENTERPRISE': 'Enterprise edition',
            'ENTERPRISE_PLUS': 'Enterprise Plus edition',
            'EDITION_UNSPECIFIED': (
                "Spanner's legacy pricing model. For more information, see the"
                ' [Spanner editions overview]'
                '(https://cloud.google.com/spanner/docs/editions-overview)'
            ),
        },
    ).AddToParser(parser)
    flags.DefaultBackupScheduleType(
        choices={
            'DEFAULT_BACKUP_SCHEDULE_TYPE_UNSPECIFIED': 'Not specified.',
            'NONE': (
                'No default backup schedule is created automatically when a new'
                ' database is created in an instance.'
            ),
            'AUTOMATIC': (
                'A default backup schedule is created automatically when a new'
                ' database is created in an instance. You can edit or delete'
                " the default backup schedule once it's created. The default"
                ' backup schedule creates a full backup every 24 hours. These'
                ' full backups are retained for 7 days.'
            ),
        },
    ).AddToParser(parser)
    resource_args.AddExpireBehaviorArg(parser)
    resource_args.AddInstanceTypeArg(parser)
    flags.AddCapacityArgsForInstance(
        require_all_autoscaling_args=True,
        parser=parser,
        add_asymmetric_option_flag=True,
        autoscaling_cpu_target_group=True,
        add_asymmetric_total_cpu_target_flag=True,
        add_asymmetric_disable_autoscaling_flags=True,
        add_disable_downscaling_flag=True,
    )
    base.ASYNC_FLAG.AddToParser(parser)
    parser.display_info.AddCacheUpdater(flags.InstanceCompleter)
    flags.AddTags(parser)
    labels_util.AddCreateLabelsFlags(parser)

  def Run(self, args):
    """This is what gets called when the user runs this command.

    Args:
      args: an argparse namespace. All the arguments that were provided to this
        command invocation.

    Returns:
      Some value that we want to have printed later.
    """
    instance_type = resource_args.GetInstanceType(args)
    expire_behavior = resource_args.GetExpireBehavior(args)

    op = instances.Create(
        instance=args.instance,
        config=args.config,
        description=args.description,
        nodes=args.nodes,
        processing_units=args.processing_units,
        autoscaling_min_nodes=args.autoscaling_min_nodes,
        autoscaling_max_nodes=args.autoscaling_max_nodes,
        autoscaling_min_processing_units=args.autoscaling_min_processing_units,
        autoscaling_max_processing_units=args.autoscaling_max_processing_units,
        autoscaling_high_priority_cpu_target=args.autoscaling_high_priority_cpu_target,
        autoscaling_total_cpu_target=args.autoscaling_total_cpu_target,
        autoscaling_storage_target=args.autoscaling_storage_target,
        asymmetric_autoscaling_options=args.asymmetric_autoscaling_option,
        disable_downscaling=args.disable_downscaling,
        instance_type=instance_type,
        expire_behavior=expire_behavior,
        edition=args.edition,
        default_backup_schedule_type=args.default_backup_schedule_type,
        tags=args.tags,
        labels=args.labels,
    )
    if args.async_:
      return op
    instance_operations.Await(op, 'Creating instance')


@base.UniverseCompatible
@base.ReleaseTracks(base.ReleaseTrack.ALPHA)
class AlphaCreate(Create):
  """Create a Cloud Spanner instance with ALPHA features."""
  __doc__ = Create.__doc__

  @staticmethod
  def Args(parser):
    """See base class."""
    flags.Instance().AddToParser(parser)
    flags.Config().AddToParser(parser)
    flags.Description().AddToParser(parser)
    flags.SsdCache().AddToParser(parser)
    flags.Edition(
        choices={
            'STANDARD': 'Standard edition',
            'ENTERPRISE': 'Enterprise edition',
            'ENTERPRISE_PLUS': 'Enterprise Plus edition',
            'EDITION_UNSPECIFIED': (
                "Spanner's legacy pricing model. For more information, see the"
                ' [Spanner editions overview]'
                '(https://cloud.google.com/spanner/docs/editions-overview)'
            ),
        },
    ).AddToParser(parser)
    flags.DefaultBackupScheduleType(
        choices={
            'DEFAULT_BACKUP_SCHEDULE_TYPE_UNSPECIFIED': 'Not specified.',
            'NONE': (
                'No default backup schedule is created automatically when a new'
                ' database is created in an instance.'
            ),
            'AUTOMATIC': (
                'A default backup schedule is created automatically when a new'
                ' database is created in an instance. You can edit or delete'
                " the default backup schedule once it's created. The default"
                ' backup schedule creates a full backup every 24 hours. These'
                ' full backups are retained for 7 days.'
            ),
        },
    ).AddToParser(parser)
    resource_args.AddExpireBehaviorArg(parser)
    resource_args.AddInstanceTypeArg(parser)
    flags.AddCapacityArgsForInstance(
        require_all_autoscaling_args=True,
        parser=parser,
        add_asymmetric_option_flag=True,
        asymmetric_options_group=False,
        autoscaling_cpu_target_group=True,
        add_asymmetric_total_cpu_target_flag=True,
        add_asymmetric_disable_autoscaling_flags=True,
        add_disable_downscaling_flag=True,
    )
    base.ASYNC_FLAG.AddToParser(parser)
    parser.display_info.AddCacheUpdater(flags.InstanceCompleter)
    flags.AddTags(parser)
    labels_util.AddCreateLabelsFlags(parser)

  def Run(self, args):
    """This is what gets called when the user runs this command.

    Args:
      args: an argparse namespace. All the arguments that were provided to this
        command invocation.

    Returns:
      Some value that we want to have printed later.
    """
    instance_type = resource_args.GetInstanceType(args)
    expire_behavior = resource_args.GetExpireBehavior(args)

    op = instances.Create(
        instance=args.instance,
        config=args.config,
        description=args.description,
        nodes=args.nodes,
        processing_units=args.processing_units,
        autoscaling_min_nodes=args.autoscaling_min_nodes,
        autoscaling_max_nodes=args.autoscaling_max_nodes,
        autoscaling_min_processing_units=args.autoscaling_min_processing_units,
        autoscaling_max_processing_units=args.autoscaling_max_processing_units,
        autoscaling_high_priority_cpu_target=args.autoscaling_high_priority_cpu_target,
        autoscaling_total_cpu_target=args.autoscaling_total_cpu_target,
        autoscaling_storage_target=args.autoscaling_storage_target,
        asymmetric_autoscaling_options=args.asymmetric_autoscaling_option,
        disable_downscaling=args.disable_downscaling,
        instance_type=instance_type,
        expire_behavior=expire_behavior,
        ssd_cache=args.ssd_cache,
        edition=args.edition,
        default_backup_schedule_type=args.default_backup_schedule_type,
        tags=args.tags,
        labels=args.labels,
    )
    if args.async_:
      return op
    instance_operations.Await(op, 'Creating instance')



# -*- coding: utf-8 -*- #
# Copyright 2022 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Command for creating machine images."""


import uuid

from googlecloudsdk.api_lib.compute import base_classes
from googlecloudsdk.calliope import arg_parsers
from googlecloudsdk.calliope import base
from googlecloudsdk.command_lib.compute import flags as compute_flags
from googlecloudsdk.command_lib.compute import scope as compute_scopes
from googlecloudsdk.command_lib.compute.instances import flags as instances_flags
from googlecloudsdk.command_lib.compute.instances.bulk import flags as bulk_flags
from googlecloudsdk.command_lib.compute.instances.bulk import util as bulk_util
from googlecloudsdk.command_lib.compute.queued_resources import flags as queued_resource_flags
from googlecloudsdk.core import log


@base.UniverseCompatible
class Create(base.CreateCommand):
  """Create a Compute Engine queued resource."""

  _ALLOW_RSA_ENCRYPTED_CSEK_KEYS = True

  detailed_help = {
      'brief': 'Create a Compute Engine queued resource.',
      'EXAMPLES': """
     To create a queued resource, run:

       $ {command} queued-resource-1 --count=1 --name-pattern=instance-#
         --valid-until-duration=7d --zone=us-central1-a
   """,
  }
  # LINT.IfChange(alpha_spec)
  _support_display_device = True
  _support_secure_tags = True
  _support_numa_node_count = True
  _support_snp_svsm = True
  _support_max_count_per_zone = False
  _support_custom_hostnames = True
  _support_specific_then_x = True
  _support_any_reservation_then_fail_affinity = True
  _support_watchdog_timer = True
  _support_igmp_query = True
  _support_graceful_shutdown = True
  _support_flex_start = True
  _support_source_snapshot_region = True
  _support_skip_guest_os_shutdown = True
  _support_preemption_notice_duration = True
  _support_workload_identity_config = True
  _support_instance_selection_min_cpu_platform = True
  _support_vsock_mode = True

  @classmethod
  def Args(cls, parser):
    # Bulk insert specific flags
    bulk_flags.AddCommonBulkInsertArgs(
        parser,
        base.ReleaseTrack.ALPHA,
        support_display_device=cls._support_display_device,
        support_numa_node_count=cls._support_numa_node_count,
        add_zone_region_flags=False,
        support_snp_svsm=cls._support_snp_svsm,
        support_max_count_per_zone=cls._support_max_count_per_zone,
        support_custom_hostnames=cls._support_custom_hostnames,
        support_specific_then_x_affinity=cls._support_specific_then_x,
        support_any_reservation_then_fail_affinity=cls._support_any_reservation_then_fail_affinity,
        support_watchdog_timer=cls._support_watchdog_timer,
        support_igmp_query=cls._support_igmp_query,
        support_graceful_shutdown=cls._support_graceful_shutdown,
        support_flex_start=cls._support_flex_start,
        support_source_snapshot_region=cls._support_source_snapshot_region,
        support_skip_guest_os_shutdown=cls._support_skip_guest_os_shutdown,
        support_preemption_notice_duration=cls._support_preemption_notice_duration,
        support_instance_flexibility_policy=False,
        support_workload_identity_config=cls._support_workload_identity_config,
        support_instance_selection_min_cpu_platform=cls._support_instance_selection_min_cpu_platform,
        support_vsock_mode=cls._support_vsock_mode,
    )
    cls.AddSourceInstanceTemplate(parser)
    instances_flags.AddSecureTagsArgs(parser)
    instances_flags.AddHostErrorTimeoutSecondsArgs(parser)
    instances_flags.AddLocalSsdRecoveryTimeoutArgs(parser)
    instances_flags.AddMaintenanceInterval().AddToParser(parser)
    # LINT.ThenChange(../instances/bulk/create.py:alpha_spec)
    # Queued resource specific flags
    valid_until_group = parser.add_group(mutex=True, required=True)
    valid_until_group.add_argument(
        '--valid-until-duration',
        type=arg_parsers.Duration(),
        help="""Relative deadline for waiting for capacity.""",
    )
    valid_until_group.add_argument(
        '--valid-until-time',
        type=arg_parsers.Datetime.Parse,
        help="""Absolute deadline for waiting for capacity in RFC3339 text format.""",
    )
    Create.QueuedResourceArg = queued_resource_flags.MakeQueuedResourcesArg(
        plural=False
    )
    Create.QueuedResourceArg.AddArgument(parser, operation_type='create')
    queued_resource_flags.AddOutputFormat(parser)

  # LINT.IfChange(instance_template)
  @classmethod
  def AddSourceInstanceTemplate(cls, parser):
    cls.SOURCE_INSTANCE_TEMPLATE = (
        bulk_flags.MakeBulkSourceInstanceTemplateArg()
    )
    cls.SOURCE_INSTANCE_TEMPLATE.AddArgument(parser)

  # LINT.ThenChange(../instances/bulk/create.py:instance_template)

  def Run(self, args):
    bulk_flags.ValidateBulkInsertArgs(
        args,
        support_max_count_per_zone=self._support_max_count_per_zone,
        support_custom_hostnames=self._support_custom_hostnames,
        support_instance_flexibility_policy=False,
    )

    holder = base_classes.ComputeApiHolder(self.ReleaseTrack())
    client = holder.client

    queued_resource_ref = Create.QueuedResourceArg.ResolveAsResource(
        args,
        holder.resources,
        scope_lister=compute_flags.GetDefaultScopeLister(client),
    )

    zone = args.zone
    if not zone and queued_resource_ref.zone:
      zone = queued_resource_ref.zone

    supported_features = bulk_util.SupportedFeatures(
        self._support_display_device,
        self._support_secure_tags,
        self._support_numa_node_count,
        self._support_snp_svsm,
        self._support_max_count_per_zone,
        self._support_custom_hostnames,
        self._support_specific_then_x,
        self._support_any_reservation_then_fail_affinity,
        self._support_watchdog_timer,
        self._support_graceful_shutdown,
        self._support_source_snapshot_region,
        self._support_skip_guest_os_shutdown,
        self._support_preemption_notice_duration,
        # Instance flexibility policy is not supported in QueuedResource.
        False,
        self._support_workload_identity_config,
        self._support_instance_selection_min_cpu_platform,
        support_vsock_mode=self._support_vsock_mode,
    )
    bulk_insert_instance_resource = bulk_util.CreateBulkInsertInstanceResource(
        args,
        holder,
        client,
        holder.resources,
        queued_resource_ref.project,
        zone,
        compute_scopes.ScopeEnum.ZONE,
        self.SOURCE_INSTANCE_TEMPLATE,
        supported_features,
    )

    # minCount is not supported in QueuedResource
    bulk_insert_instance_resource.reset('minCount')

    queued_resource = client.messages.QueuedResource(
        name=queued_resource_ref.Name(),
        queuingPolicy=client.messages.QueuingPolicy(
            validUntilDuration=client.messages.Duration(
                seconds=args.valid_until_duration
            )
        ),
        bulkInsertInstanceResource=bulk_insert_instance_resource,
    )

    request = client.messages.ComputeZoneQueuedResourcesInsertRequest(
        queuedResource=queued_resource,
        project=queued_resource_ref.project,
        zone=queued_resource_ref.zone,
        requestId=str(uuid.uuid4()),
    )
    if args.async_:
      response = client.apitools_client.zoneQueuedResources.Insert(request)
      log.status.Print(
          'Queued resource creation in progress: {}'.format(response.selfLink)
      )
      # Disable argument formatting since we have not created a resource yet.
      args.format = 'disable'
      return response
    return client.MakeRequests(
        [(client.apitools_client.zoneQueuedResources, 'Insert', request)]
    )



# -*- coding: utf-8 -*- #
# Copyright 2017 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A command that regenerates existing or new gcloud API."""


import collections
import fnmatch
import os
import re
import shutil

import googlecloudsdk
from googlecloudsdk import third_party
from googlecloudsdk.api_lib.regen import generate
from googlecloudsdk.calliope import arg_parsers
from googlecloudsdk.calliope import base
from googlecloudsdk.calliope import parser_errors
from googlecloudsdk.command_lib.meta import regen as regen_utils
from googlecloudsdk.core import log
from googlecloudsdk.core.util import encoding
from googlecloudsdk.core.util import files

import ruamel.yaml

_API_REGEX = '([a-z0-9_]+)/([a-z0-9_]+)'


@base.UniverseCompatible
class Regen(base.Command):
  """Regenerate given API(s) in gcloud."""

  @staticmethod
  def Args(parser):
    parser.add_argument(
        'api',
        type=arg_parsers.ArgList(),
        help='The APIs to regenerate in api_name/api_version format. '
             'These can be filename glob expressions to regenerate multiple '
             'apis. For example */* to regegenerate all apis and versions, '
             'or */*beta* to only regenerate existing beta apis. '
             'Note that if discovery doc is supplied this cannot '
             'contain any wildcards.')

    parser.add_argument(
        '--api-discovery-doc',
        help='Path to json file describing the api. If not specified, '
             'an existing discovery doc will be used.')

    parser.add_argument('--config',
                        help='Regeneration config yaml filename. '
                             'If not specified canonical config will be used.')

    parser.add_argument(
        '--base-dir',
        help='Directory where generated code will be written. '
        'By default googlecloudsdk/generated_clients/apis.')

  def Run(self, args):
    config = _LoadConfig(args.config)
    root_dir = config['root_dir']
    changed_config = False

    if args.api_discovery_doc:
      if not os.path.isfile(args.api_discovery_doc):
        raise regen_utils.DiscoveryDocError(
            f'File not found {args.api_discovery_doc}')
      if len(args.api) != 1:
        raise parser_errors.ArgumentError(
            'Can only specify one api when discovery doc is provided.')

      match = re.match(_API_REGEX, args.api[0])
      if not match:
        raise regen_utils.DiscoveryDocError(
            f'Api name must match {_API_REGEX} pattern when discovery doc '
            'is specified')

      api_name, api_version = match.group(1), match.group(2)
      if api_name not in config['apis']:
        log.warning('No such api %s in config, adding...', api_name)
        config['apis'][api_name] = {api_version: {'discovery_doc': ''}}
        changed_config = True
      elif api_version not in config['apis'][api_name]:
        log.warning('No such api version %s in config, adding...', api_version)
        config['apis'][api_name][api_version] = {'discovery_doc': ''}
        changed_config = True

      api_version_config = config['apis'].get(api_name).get(api_version, {})
      discovery_doc = api_name + '_' + api_version + '.json'
      new_discovery_doc = os.path.realpath(args.api_discovery_doc)
      old_discovery_doc = os.path.realpath(
          os.path.join(args.base_dir, root_dir, discovery_doc))

      if new_discovery_doc != old_discovery_doc:
        log.status.Print(f'Copying in {new_discovery_doc}')
        shutil.copyfile(new_discovery_doc, old_discovery_doc)

      if api_version_config['discovery_doc'] != discovery_doc:
        changed_config = True
        api_version_config['discovery_doc'] = discovery_doc

      regenerate_list = [
          (match.group(1), match.group(2), api_version_config)
      ]
    else:
      regex_patern = '|'.join(map(fnmatch.translate, args.api))
      regenerate_list = []
      for api_name, api_version_config in config['apis'].items():
        for api_version, api_config in api_version_config.items():
          if re.match(regex_patern, api_name + '/' + api_version):
            regenerate_list.append((api_name, api_version, api_config))

    if not regenerate_list:
      raise regen_utils.UnknownApi(
          f'api [{",".join(args.api)}] not found in "apis" section of '
          f'{args.config}. Use [gcloud meta apis list] to see available apis.')

    base_dir = args.base_dir or os.path.dirname(
        os.path.dirname(googlecloudsdk.__file__))
    for api_name, api_version, api_config in sorted(regenerate_list):
      log.status.Print(
          f'Generating {api_name} {api_version} from '
          f'{os.path.join(root_dir, api_config["discovery_doc"])}')
      discovery_doc = os.path.join(
          base_dir, root_dir, api_config['discovery_doc'])
      output_dir = os.path.join(base_dir, root_dir)
      root_package = root_dir.replace('/', '.')

      generate.GenerateApitoolsApi(
          discovery_doc, output_dir, root_package, api_name, api_version,
          api_config)
      generate.GenerateApitoolsResourceModule(
          discovery_doc,
          output_dir,
          api_name,
          api_version,
          api_config.get('resources', {}),
      )

    client_output_dir = os.path.join(base_dir, root_dir)
    apis_config_map = collections.defaultdict(dict)
    for api_name, versions_config in config['apis'].items():
      for version, version_config in versions_config.items():
        apis_config_map[api_name][version] = generate.ApiConfig.FromData(
            data=version_config,
            root_dir=root_dir,
            discovery_doc_dir=client_output_dir,
        )
    generate.GenerateApiMap(
        os.path.join(client_output_dir, 'apis_map.py'), apis_config_map)

    # Now that everything passed, config can be updated if needed.
    if changed_config:
      log.warning('Updated %s', args.config)
      with files.FileWriter(args.config) as stream:
        ruamel.yaml.round_trip_dump(config, stream)


def _LoadConfig(config_file_name=None):
  """Loads regen config from given filename."""
  config_file_name = config_file_name or os.path.join(
      os.path.dirname(encoding.Decode(third_party.__file__)),
      'regen_apis_config.yaml')

  if not os.path.isfile(config_file_name):
    raise regen_utils.ConfigFileError(f'{config_file_name} Not found')
  with files.FileReader(config_file_name) as stream:
    config = ruamel.yaml.round_trip_load(stream)
  if not config or 'root_dir' not in config:
    raise regen_utils.ConfigFileError(
        f'{config_file_name} does not have format of gcloud api config file')
  return config



# -*- coding: utf-8 -*- #
# Copyright 2025 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Updates a new Backup Plan."""


import argparse

from apitools.base.py import exceptions as apitools_exceptions
from googlecloudsdk.api_lib.backupdr import backup_plans
from googlecloudsdk.api_lib.backupdr import util
from googlecloudsdk.api_lib.util import exceptions
from googlecloudsdk.calliope import base
from googlecloudsdk.command_lib.backupdr import flags
from googlecloudsdk.core import exceptions as core_exceptions
from googlecloudsdk.core import log
from googlecloudsdk.core import yaml


@base.UniverseCompatible
@base.ReleaseTracks(
    base.ReleaseTrack.ALPHA, base.ReleaseTrack.BETA, base.ReleaseTrack.GA
)
class Update(base.UpdateCommand):
  """Update a specific backup plan."""

  detailed_help = {
      'BRIEF': 'Update a specific backup plan',
      'DESCRIPTION': """\
          Update a specific backup plan in the project. It can only be updated in regions supported by the Backup and DR Service.
      """,
      'EXAMPLES': """\
        To update 2 backup rules and description of an existing backup plan ``sample-backup-plan''
        in project ``sample-project'',
        at location ``us-central1'':

        run:

          $ {command} sample-backup-plan --project=sample-project --location=us-central1
            --backup-rule <BACKUP-RULE>
            --backup-rule <BACKUP-RULE>
            --description "This is a sample backup plan"

        To add backup rules to an existing backup plan ``sample-backup-plan''
        in project ``sample-project'',
        at location ``us-central1'':

        run:

          $ {command} sample-backup-plan --project=sample-project --location=us-central1
            --add-backup-rule <BACKUP-RULE>
            --add-backup-rule <BACKUP-RULE>

        To remove a backup rule with id ``sample-daily-rule'' from an existing backup plan ``sample-backup-plan''
        in project ``sample-project'',
        at location ``us-central1'':

        run:

          $ {command} sample-backup-plan --project=sample-project --location=us-central1
            --remove-backup-rule sample-daily-rule

        To override backup rules in an existing backup plan ``sample-backup-plan''
        in project ``sample-project'',
        at location ``us-central1'', pass a file path containing backup rules in YAML or JSON format:
        This flag is mutually exclusive with --add-backup-rule, --remove-backup-rule and --backup-rule flags.

        run:
          $ {command} sample-backup-plan --project=sample-project --location=us-central1
            --backup-rules-fom-file <FILE_PATH>

        Backup Rule Examples:

        1. Hourly backup rule with hourly backup frequency of 6 hours and store it for 30 days, and expect the backups to run only between 10:00 to 20:00 UTC

        <BACKUP-RULE>: rule-id=sample-hourly-rule,retention-days=30,recurrence=HOURLY,hourly-frequency=6,time-zone=UTC,backup-window-start=10,backup-window-end=20

        Properties:
          -- rule-id = "sample-hourly-rule"
          -- retention-days = 30
          -- recurrence = HOURLY
          -- hourly-frequency = 6
          -- time-zone = UTC
          -- backup-window-start = 10
          -- backup-window-end = 20

        2. Daily backup rule with daily backup frequency of 6 hours and store it for 7 days

        <BACKUP-RULE>: rule-id=sample-daily-rule,retention-days=7,recurrence=DAILY,backup-window-start=1,backup-window-end=14

        Properties:
          -- rule-id = "sample-daily-rule"
          -- retention-days = 7
          -- recurrence = DAILY
          -- backup-window-start = 1
          -- backup-window-end = 14

        3. Weekly backup rule with weekly backup frequency on every MONDAY & FRIDAY and store it for 21 days

        <BACKUP-RULE>: rule-id=sample-weekly-rule,retention-days=21,recurrence=WEEKLY,days-of-week="MONDAY FRIDAY",backup-window-start=10,backup-window-end=20

        Properties:
          -- rule-id = "sample-weekly-rule"
          -- retention-days: 21
          -- recurrence = WEEKLY
          -- days-of-week = "MONDAY FRIDAY"
          -- backup-window-start = 10
          -- backup-window-end = 20

        YAML and JSON file examples:

        YAML file example:

        ```
        backup-rules:
        - rule-id: weekly-rule
          retention-days: 7
          recurrence: WEEKLY
          backup-window-start: 0
          backup-window-end: 23
          days-of-week: [MONDAY, TUESDAY]
          time-zone: UTC
        - rule-id: daily-rule
          retention-days: 1
          recurrence: DAILY
          backup-window-start: 1
          backup-window-end: 24
          time-zone: UTC
        ```

        JSON file example:
        ```
        {
          "backup-rules": [
            {
              "rule-id": "weekly-rule",
              "retention-days": 7,
              "recurrence": "WEEKLY",
              "backup-window-start": 0,
              "backup-window-end": 23,
              "days-of-week": ["MONDAY", "TUESDAY"],
              "time-zone": "UTC"
            },
            {
              "rule-id": "daily-rule",
              "retention-days": 1,
              "recurrence": "DAILY",
              "backup-window-start": 1,
              "backup-window-end": 24,
              "time-zone": "UTC"
            }
          ]
        }
        ```
        """,
  }

  @staticmethod
  def Args(parser: argparse.ArgumentParser):
    """Specifies additional command flags.

    Args:
      parser: Parser object for command line inputs.
    """
    base.ASYNC_FLAG.AddToParser(parser)
    base.ASYNC_FLAG.SetDefault(parser, True)
    flags.AddBackupPlanResourceArg(
        parser,
        """Name of the backup plan to be updated.
        The name must be unique for a project and location.
        """,
    )
    flags.AddUpdateBackupRule(parser)
    flags.AddAddBackupRule(parser)
    flags.AddRemoveBackupRule(parser)
    flags.AddBackupRulesFromFile(parser)
    flags.AddDiskBackupPlanProperties(parser)
    flags.AddComputeInstanceBackupPlanProperties(parser)

    description_help = """\
        Provide a description of the backup plan, such as specific use cases and
        relevant details, in 2048 characters or less.

        E.g., This is a backup plan that performs a daily backup at 6 p.m. and
        retains data for 3 months.
        """
    flags.AddDescription(parser, description_help)
    flags.AddLogRetentionDays(parser)
    flags.AddMaxCustomOnDemandRetentionDays(parser)

  class YamlOrJsonLoadError(core_exceptions.Error):
    """Error parsing YAML or JSON file content."""

  def _GetBackupRulesFromFile(self, backup_rules_file_content):
    """Parses the backup rules from the file content."""
    try:
      backup_rules = yaml.load(backup_rules_file_content)
      return backup_rules.get('backup-rules')
    except Exception as exc:  # pylint: disable=broad-except
      raise self.YamlOrJsonLoadError(
          'Could not parse content in the backup rules file: {0}'.format(exc)
      ) from exc

  def Run(self, args) -> any:
    """Constructs and sends request.

    Args:
      args: argparse.Namespace, An object that contains the values for the
        arguments specified in the .Args() method.

    Returns:
      ProcessHttpResponse of the request made.
    """
    api_version = util.GetApiVersion(self.ReleaseTrack())
    client = backup_plans.BackupPlansClient(api_version=api_version)

    backup_plan = args.CONCEPTS.backup_plan.Parse()
    backup_rules_file_content = args.backup_rules_from_file
    update_backup_rules = args.backup_rule
    add_backup_rules = args.add_backup_rule
    remove_backup_rules = args.remove_backup_rule
    disk_properties = args.disk_properties
    compute_instance_properties = args.compute_instance_properties

    if backup_rules_file_content and (
        update_backup_rules or add_backup_rules or remove_backup_rules
    ):
      raise core_exceptions.Error(
          '--backup-rules-from-file flag cannot be used with'
          ' --backup-rule, --add-backup-rule or --remove-backup-rule flags.'
      )

    log_retention_days = args.log_retention_days
    description = args.description
    max_custom_on_demand_retention_days = (
        args.max_custom_on_demand_retention_days
    )

    try:
      current_backup_plan = client.Describe(backup_plan)
      new_backup_rules_from_file = None
      if backup_rules_file_content:
        new_backup_rules_from_file = self._GetBackupRulesFromFile(
            backup_rules_file_content
        )
      updated_backup_plan = client.ParseUpdate(
          description,
          new_backup_rules_from_file,
          update_backup_rules,
          add_backup_rules,
          remove_backup_rules,
          current_backup_plan,
          log_retention_days,
          max_custom_on_demand_retention_days,
          disk_properties,
          compute_instance_properties,
      )
      update_mask = []
      if (
          description is not None
          and description != current_backup_plan.description
      ):
        update_mask.append('description')
      if (
          log_retention_days is not None
          and log_retention_days != current_backup_plan.logRetentionDays
      ):
        update_mask.append('logRetentionDays')
      if (
          max_custom_on_demand_retention_days is not None
          and max_custom_on_demand_retention_days
          != current_backup_plan.maxCustomOnDemandRetentionDays
      ):
        update_mask.append('maxCustomOnDemandRetentionDays')
      if args.IsSpecified('disk_properties'):
        update_mask.append('disk_backup_plan_properties.guest_flush')
      if args.IsSpecified('compute_instance_properties'):
        update_mask.append(
            'compute_instance_backup_plan_properties.guest_flush'
        )
      if any([
          update_backup_rules,
          add_backup_rules,
          remove_backup_rules,
          new_backup_rules_from_file,
      ]):
        update_mask.append('backupRules')
      operation = client.Update(
          backup_plan, updated_backup_plan, ','.join(update_mask)
      )
    except apitools_exceptions.HttpError as e:
      raise exceptions.HttpException(e, util.HTTP_ERROR_FORMAT)

    if args.async_:
      log.UpdatedResource(
          backup_plan.RelativeName(),
          kind='backup plan',
          is_async=True,
          details=util.ASYNC_OPERATION_MESSAGE.format(operation.name),
      )
      return operation

    resource = client.WaitForOperation(
        operation_ref=client.GetOperationRef(operation),
        message=(
            f'Updating backup plan [{backup_plan.RelativeName()}]. (This'
            ' operation could take up to 2 minutes.)'
        ),
    )
    log.UpdatedResource(backup_plan.RelativeName(), kind='backup plan')

    return resource
